# Inspiration: https://github.com/apify/crawlee/blob/v3.9.2/packages/utils/src/internals/error_tracker.ts

from __future__ import annotations

from collections import Counter
from dataclasses import dataclass


@dataclass(frozen=True, unsafe_hash=True)
class ErrorGroup:
    """Identifies a group of similar errors."""

    class_name: str | None


class ErrorTracker:
    """Track errors and aggregates their counts by similarity."""

    def __init__(self) -> None:
        self._errors = Counter[ErrorGroup]()

    def add(self, error: Exception) -> None:
        """Include an error in the statistics."""
        error_group = ErrorGroup(class_name=error.__class__.__name__)
        self._errors[error_group] += 1

    @property
    def unique_error_count(self) -> int:
        """Number of distinct kinds of errors."""
        return len(self._errors)

    @property
    def total(self) -> int:
        """Total number of errors."""
        return sum(self._errors.values())



from .base_http_client import BaseHttpClient, HttpCrawlingResult, HttpResponse
from .httpx_client import HttpxClient

__all__ = ['BaseHttpClient', 'HttpCrawlingResult', 'HttpResponse', 'HttpxClient']



from __future__ import annotations

import asyncio
import json
import os
from datetime import datetime, timezone
from logging import getLogger
from typing import TYPE_CHECKING, Any, AsyncContextManager, AsyncIterator

import aiofiles
import aioshutil
from aiofiles.os import makedirs
from typing_extensions import override

from crawlee._utils.crypto import crypto_random_object_id
from crawlee._utils.data_processing import raise_on_duplicate_storage, raise_on_non_existing_storage
from crawlee._utils.file import force_rename, json_dumps
from crawlee.base_storage_client import BaseDatasetClient
from crawlee.memory_storage_client._creation_management import find_or_create_client_by_id_or_name_inner
from crawlee.models import DatasetItemsListPage, DatasetMetadata
from crawlee.types import StorageTypes

if TYPE_CHECKING:
    from httpx import Response

    from crawlee.memory_storage_client import MemoryStorageClient
    from crawlee.types import JSONSerializable

logger = getLogger(__name__)


class DatasetClient(BaseDatasetClient):
    """Subclient for manipulating a single dataset."""

    _LIST_ITEMS_LIMIT = 999_999_999_999
    """This is what API returns in the x-apify-pagination-limit header when no limit query parameter is used."""

    _LOCAL_ENTRY_NAME_DIGITS = 9
    """Number of characters of the dataset item file names, e.g.: 000000019.json - 9 digits."""

    def __init__(
        self,
        *,
        memory_storage_client: MemoryStorageClient,
        id: str | None = None,
        name: str | None = None,
        created_at: datetime | None = None,
        accessed_at: datetime | None = None,
        modified_at: datetime | None = None,
        item_count: int = 0,
    ) -> None:
        self._memory_storage_client = memory_storage_client
        self.id = id or crypto_random_object_id()
        self.name = name
        self._created_at = created_at or datetime.now(timezone.utc)
        self._accessed_at = accessed_at or datetime.now(timezone.utc)
        self._modified_at = modified_at or datetime.now(timezone.utc)

        self.dataset_entries: dict[str, dict] = {}
        self.file_operation_lock = asyncio.Lock()
        self.item_count = item_count

    @property
    def resource_info(self) -> DatasetMetadata:
        """Get the resource info for the dataset client."""
        return DatasetMetadata(
            id=self.id,
            name=self.name,
            accessed_at=self._accessed_at,
            created_at=self._created_at,
            modified_at=self._modified_at,
            item_count=self.item_count,
        )

    @property
    def resource_directory(self) -> str:
        """Get the resource directory for the client."""
        return os.path.join(self._memory_storage_client.datasets_directory, self.name or self.id)

    @override
    async def get(self) -> DatasetMetadata | None:
        found = find_or_create_client_by_id_or_name_inner(
            resource_client_class=DatasetClient,
            memory_storage_client=self._memory_storage_client,
            id=self.id,
            name=self.name,
        )

        if found:
            async with found.file_operation_lock:
                await found.update_timestamps(has_been_modified=False)
                return found.resource_info

        return None

    @override
    async def update(self, *, name: str | None = None) -> DatasetMetadata:
        # Check by id
        existing_dataset_by_id = find_or_create_client_by_id_or_name_inner(
            resource_client_class=DatasetClient,
            memory_storage_client=self._memory_storage_client,
            id=self.id,
            name=self.name,
        )

        if existing_dataset_by_id is None:
            raise_on_non_existing_storage(StorageTypes.DATASET, self.id)

        # Skip if no changes
        if name is None:
            return existing_dataset_by_id.resource_info

        async with existing_dataset_by_id.file_operation_lock:
            # Check that name is not in use already
            existing_dataset_by_name = next(
                (
                    dataset
                    for dataset in self._memory_storage_client.datasets_handled
                    if dataset.name and dataset.name.lower() == name.lower()
                ),
                None,
            )

            if existing_dataset_by_name is not None:
                raise_on_duplicate_storage(StorageTypes.DATASET, 'name', name)

            previous_dir = existing_dataset_by_id.resource_directory
            existing_dataset_by_id.name = name

            await force_rename(previous_dir, existing_dataset_by_id.resource_directory)

            # Update timestamps
            await existing_dataset_by_id.update_timestamps(has_been_modified=True)

        return existing_dataset_by_id.resource_info

    @override
    async def delete(self) -> None:
        dataset = next(
            (dataset for dataset in self._memory_storage_client.datasets_handled if dataset.id == self.id), None
        )

        if dataset is not None:
            async with dataset.file_operation_lock:
                self._memory_storage_client.datasets_handled.remove(dataset)
                dataset.item_count = 0
                dataset.dataset_entries.clear()

                if os.path.exists(dataset.resource_directory):
                    await aioshutil.rmtree(dataset.resource_directory)

    @override
    async def list_items(
        self,
        *,
        offset: int | None = 0,
        limit: int | None = _LIST_ITEMS_LIMIT,
        clean: bool = False,
        desc: bool = False,
        fields: list[str] | None = None,
        omit: list[str] | None = None,
        unwind: str | None = None,
        skip_empty: bool = False,
        skip_hidden: bool = False,
        flatten: list[str] | None = None,
        view: str | None = None,
    ) -> DatasetItemsListPage:
        # Check by id
        existing_dataset_by_id = find_or_create_client_by_id_or_name_inner(
            resource_client_class=DatasetClient,
            memory_storage_client=self._memory_storage_client,
            id=self.id,
            name=self.name,
        )

        if existing_dataset_by_id is None:
            raise_on_non_existing_storage(StorageTypes.DATASET, self.id)

        async with existing_dataset_by_id.file_operation_lock:
            start, end = existing_dataset_by_id.get_start_and_end_indexes(
                max(existing_dataset_by_id.item_count - (offset or 0) - (limit or self._LIST_ITEMS_LIMIT), 0)
                if desc
                else offset or 0,
                limit,
            )

            items = []

            for idx in range(start, end):
                entry_number = self._generate_local_entry_name(idx)
                items.append(existing_dataset_by_id.dataset_entries[entry_number])

            await existing_dataset_by_id.update_timestamps(has_been_modified=False)

            if desc:
                items.reverse()

            return DatasetItemsListPage(
                count=len(items),
                desc=desc or False,
                items=items,
                limit=limit or self._LIST_ITEMS_LIMIT,
                offset=offset or 0,
                total=existing_dataset_by_id.item_count,
            )

    @override
    async def iterate_items(  # type: ignore
        self,
        *,
        offset: int = 0,
        limit: int | None = None,
        clean: bool = False,
        desc: bool = False,
        fields: list[str] | None = None,
        omit: list[str] | None = None,
        unwind: str | None = None,
        skip_empty: bool = False,
        skip_hidden: bool = False,
    ) -> AsyncIterator[dict]:
        cache_size = 1000
        first_item = offset

        # If there is no limit, set last_item to None until we get the total from the first API response
        last_item = None if limit is None else offset + limit
        current_offset = first_item

        while last_item is None or current_offset < last_item:
            current_limit = cache_size if last_item is None else min(cache_size, last_item - current_offset)

            current_items_page = await self.list_items(
                offset=current_offset,
                limit=current_limit,
                desc=desc,
            )

            current_offset += current_items_page.count
            if last_item is None or current_items_page.total < last_item:
                last_item = current_items_page.total

            for item in current_items_page.items:
                yield item

    @override
    async def get_items_as_bytes(
        self,
        *,
        item_format: str = 'json',
        offset: int | None = None,
        limit: int | None = None,
        desc: bool = False,
        clean: bool = False,
        bom: bool = False,
        delimiter: str | None = None,
        fields: list[str] | None = None,
        omit: list[str] | None = None,
        unwind: str | None = None,
        skip_empty: bool = False,
        skip_header_row: bool = False,
        skip_hidden: bool = False,
        xml_root: str | None = None,
        xml_row: str | None = None,
        flatten: list[str] | None = None,
    ) -> bytes:
        raise NotImplementedError('This method is not supported in memory storage.')

    @override
    async def stream_items(
        self,
        *,
        item_format: str = 'json',
        offset: int | None = None,
        limit: int | None = None,
        desc: bool = False,
        clean: bool = False,
        bom: bool = False,
        delimiter: str | None = None,
        fields: list[str] | None = None,
        omit: list[str] | None = None,
        unwind: str | None = None,
        skip_empty: bool = False,
        skip_header_row: bool = False,
        skip_hidden: bool = False,
        xml_root: str | None = None,
        xml_row: str | None = None,
    ) -> AsyncContextManager[Response | None]:
        raise NotImplementedError('This method is not supported in memory storage.')

    @override
    async def push_items(
        self,
        items: JSONSerializable,
    ) -> None:
        # Check by id
        existing_dataset_by_id = find_or_create_client_by_id_or_name_inner(
            resource_client_class=DatasetClient,
            memory_storage_client=self._memory_storage_client,
            id=self.id,
            name=self.name,
        )

        if existing_dataset_by_id is None:
            raise_on_non_existing_storage(StorageTypes.DATASET, self.id)

        normalized = self._normalize_items(items)

        added_ids: list[str] = []
        for entry in normalized:
            existing_dataset_by_id.item_count += 1
            idx = self._generate_local_entry_name(existing_dataset_by_id.item_count)

            existing_dataset_by_id.dataset_entries[idx] = entry
            added_ids.append(idx)

        data_entries = [(id, existing_dataset_by_id.dataset_entries[id]) for id in added_ids]

        async with existing_dataset_by_id.file_operation_lock:
            await existing_dataset_by_id.update_timestamps(has_been_modified=True)

            await self._persist_dataset_items_to_disk(
                data=data_entries,
                entity_directory=existing_dataset_by_id.resource_directory,
                persist_storage=self._memory_storage_client.persist_storage,
            )

    async def _persist_dataset_items_to_disk(
        self,
        *,
        data: list[tuple[str, dict]],
        entity_directory: str,
        persist_storage: bool,
    ) -> None:
        """Writes dataset items to the disk.

        The function iterates over a list of dataset items, each represented as a tuple of an identifier
        and a dictionary, and writes them as individual JSON files in a specified directory. The function
        will skip writing if `persist_storage` is False. Before writing, it ensures that the target
        directory exists, creating it if necessary.

        Args:
            data: A list of tuples, each containing an identifier (string) and a data dictionary.
            entity_directory: The directory path where the dataset items should be stored.
            persist_storage: A boolean flag indicating whether the data should be persisted to the disk.
        """
        # Skip writing files to the disk if the client has the option set to false
        if not persist_storage:
            return

        # Ensure the directory for the entity exists
        await makedirs(entity_directory, exist_ok=True)

        # Save all the new items to the disk
        for idx, item in data:
            file_path = os.path.join(entity_directory, f'{idx}.json')
            async with aiofiles.open(file_path, mode='wb') as f:
                s = await json_dumps(item)
                await f.write(s.encode('utf-8'))

    async def update_timestamps(self, *, has_been_modified: bool) -> None:
        """Update the timestamps of the dataset."""
        from crawlee.memory_storage_client._creation_management import persist_metadata_if_enabled

        self._accessed_at = datetime.now(timezone.utc)

        if has_been_modified:
            self._modified_at = datetime.now(timezone.utc)

        await persist_metadata_if_enabled(
            data=self.resource_info.model_dump(),
            entity_directory=self.resource_directory,
            write_metadata=self._memory_storage_client.write_metadata,
        )

    def get_start_and_end_indexes(self, offset: int, limit: int | None = None) -> tuple[int, int]:
        """Calculate the start and end indexes for listing items."""
        actual_limit = limit or self.item_count
        start = offset + 1
        end = min(offset + actual_limit, self.item_count) + 1
        return (start, end)

    def _generate_local_entry_name(self, idx: int) -> str:
        return str(idx).zfill(self._LOCAL_ENTRY_NAME_DIGITS)

    def _normalize_items(self, items: JSONSerializable) -> list[dict]:
        def normalize_item(item: Any) -> dict | None:
            if isinstance(item, str):
                item = json.loads(item)

            if isinstance(item, list):
                received = ',\n'.join(item)
                raise TypeError(
                    f'Each dataset item can only be a single JSON object, not an array. Received: [{received}]'
                )

            if (not isinstance(item, dict)) and item is not None:
                raise TypeError(f'Each dataset item must be a JSON object. Received: {item}')

            return item

        if isinstance(items, str):
            items = json.loads(items)

        result = list(map(normalize_item, items)) if isinstance(items, list) else [normalize_item(items)]
        # filter(None, ..) returns items that are True
        return list(filter(None, result))



from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING, TypeVar

from crawlee.configuration import Configuration
from crawlee.memory_storage_client import MemoryStorageClient
from crawlee.storage_client_manager import StorageClientManager
from crawlee.storages import Dataset, KeyValueStore, RequestQueue

if TYPE_CHECKING:
    from crawlee.base_storage_client import BaseStorageClient
    from crawlee.base_storage_client.types import ResourceClient, ResourceCollectionClient

TResource = TypeVar('TResource', Dataset, KeyValueStore, RequestQueue)


_creation_lock = asyncio.Lock()
"""Lock for storage creation."""

_cache_dataset_by_id: dict[str, Dataset] = {}
_cache_dataset_by_name: dict[str, Dataset] = {}
_cache_kvs_by_id: dict[str, KeyValueStore] = {}
_cache_kvs_by_name: dict[str, KeyValueStore] = {}
_cache_rq_by_id: dict[str, RequestQueue] = {}
_cache_rq_by_name: dict[str, RequestQueue] = {}


def _get_from_cache_by_name(
    storage_class: type[TResource],
    name: str,
) -> TResource | None:
    """Try to restore storage from cache by name."""
    if issubclass(storage_class, Dataset):
        return _cache_dataset_by_name.get(name)  # pyright: ignore
    if issubclass(storage_class, KeyValueStore):
        return _cache_kvs_by_name.get(name)  # pyright: ignore
    if issubclass(storage_class, RequestQueue):
        return _cache_rq_by_name.get(name)  # pyright: ignore
    raise ValueError(f'Unknown storage class: {storage_class.__name__}')


def _get_from_cache_by_id(
    storage_class: type[TResource],
    id: str,
) -> TResource | None:
    """Try to restore storage from cache by ID."""
    if issubclass(storage_class, Dataset):
        return _cache_dataset_by_id.get(id)  # pyright: ignore
    if issubclass(storage_class, KeyValueStore):
        return _cache_kvs_by_id.get(id)  # pyright: ignore
    if issubclass(storage_class, RequestQueue):
        return _cache_rq_by_id.get(id)  # pyright: ignore
    raise ValueError(f'Unknown storage: {storage_class.__name__}')


def _add_to_cache_by_name(name: str, storage: TResource) -> None:
    """Add storage to cache by name."""
    if isinstance(storage, Dataset):
        _cache_dataset_by_name[name] = storage
    elif isinstance(storage, KeyValueStore):
        _cache_kvs_by_name[name] = storage
    elif isinstance(storage, RequestQueue):
        _cache_rq_by_name[name] = storage
    else:
        raise TypeError(f'Unknown storage: {storage}')


def _add_to_cache_by_id(id: str, storage: TResource) -> None:
    """Add storage to cache by ID."""
    if isinstance(storage, Dataset):
        _cache_dataset_by_id[id] = storage
    elif isinstance(storage, KeyValueStore):
        _cache_kvs_by_id[id] = storage
    elif isinstance(storage, RequestQueue):
        _cache_rq_by_id[id] = storage
    else:
        raise TypeError(f'Unknown storage: {storage}')


def _rm_from_cache_by_id(storage_class: type, id: str) -> None:
    """Remove a storage from cache by ID."""
    try:
        if issubclass(storage_class, Dataset):
            del _cache_dataset_by_id[id]
        elif issubclass(storage_class, KeyValueStore):
            del _cache_kvs_by_id[id]
        elif issubclass(storage_class, RequestQueue):
            del _cache_rq_by_id[id]
        else:
            raise TypeError(f'Unknown storage class: {storage_class.__name__}')
    except KeyError as exc:
        raise RuntimeError(f'Storage with provided ID was not found ({id}).') from exc


def _rm_from_cache_by_name(storage_class: type, name: str) -> None:
    """Remove a storage from cache by name."""
    try:
        if issubclass(storage_class, Dataset):
            del _cache_dataset_by_name[name]
        elif issubclass(storage_class, KeyValueStore):
            del _cache_kvs_by_name[name]
        elif issubclass(storage_class, RequestQueue):
            del _cache_rq_by_name[name]
        else:
            raise TypeError(f'Unknown storage class: {storage_class.__name__}')
    except KeyError as exc:
        raise RuntimeError(f'Storage with provided name was not found ({name}).') from exc


def _get_default_storage_id(configuration: Configuration, storage_class: type[TResource]) -> str:
    if issubclass(storage_class, Dataset):
        return configuration.default_dataset_id
    if issubclass(storage_class, KeyValueStore):
        return configuration.default_key_value_store_id
    if issubclass(storage_class, RequestQueue):
        return configuration.default_request_queue_id

    raise TypeError(f'Unknown storage class: {storage_class.__name__}')


async def open_storage(
    *,
    storage_class: type[TResource],
    configuration: Configuration | None = None,
    id: str | None = None,
    name: str | None = None,
) -> TResource:
    """Open either a new storage or restore an existing one and return it."""
    configuration = configuration or Configuration.get_global_configuration()
    storage_client = StorageClientManager.get_storage_client(in_cloud=configuration.in_cloud)

    # Try to restore the storage from cache by name
    if name:
        cached_storage = _get_from_cache_by_name(storage_class=storage_class, name=name)
        if cached_storage:
            return cached_storage

    default_id = _get_default_storage_id(configuration, storage_class)

    if not id and not name:
        id = default_id

    # Find out if the storage is a default on memory storage
    is_default_on_memory = id == default_id and isinstance(storage_client, MemoryStorageClient)

    # Try to restore storage from cache by ID
    if id:
        cached_storage = _get_from_cache_by_id(storage_class=storage_class, id=id)
        if cached_storage:
            return cached_storage

    # Purge on start if configured
    if configuration.purge_on_start:
        await storage_client.purge_on_start()

    # Lock and create new storage
    async with _creation_lock:
        if id and not is_default_on_memory:
            resource_client = _get_resource_client(storage_class, storage_client, id)
            storage_info = await resource_client.get()
            if not storage_info:
                raise RuntimeError(f'{storage_class.__name__} with id "{id}" does not exist!')

        elif is_default_on_memory:
            resource_collection_client = _get_resource_collection_client(storage_class, storage_client)
            storage_info = await resource_collection_client.get_or_create(name=name, id=id)

        else:
            resource_collection_client = _get_resource_collection_client(storage_class, storage_client)
            storage_info = await resource_collection_client.get_or_create(name=name)

        storage = storage_class(
            id=storage_info.id,
            name=storage_info.name,
            configuration=configuration,
            client=storage_client,
        )

        # Cache the storage by ID and name
        _add_to_cache_by_id(storage.id, storage)
        if storage.name is not None:
            _add_to_cache_by_name(storage.name, storage)

    return storage


def remove_storage_from_cache(
    *,
    storage_class: type,
    id: str | None = None,
    name: str | None = None,
) -> None:
    """Remove a storage from cache by ID or name."""
    if id:
        _rm_from_cache_by_id(storage_class=storage_class, id=id)

    if name:
        _rm_from_cache_by_name(storage_class=storage_class, name=name)


def _get_resource_client(
    storage_class: type[TResource],
    storage_client: BaseStorageClient,
    id: str,
) -> ResourceClient:
    if issubclass(storage_class, Dataset):
        return storage_client.dataset(id)

    if issubclass(storage_class, KeyValueStore):
        return storage_client.key_value_store(id)

    if issubclass(storage_class, RequestQueue):
        return storage_client.request_queue(id)

    raise ValueError(f'Unknown storage class label: {storage_class.__name__}')


def _get_resource_collection_client(
    storage_class: type,
    storage_client: BaseStorageClient,
) -> ResourceCollectionClient:
    if issubclass(storage_class, Dataset):
        return storage_client.datasets()

    if issubclass(storage_class, KeyValueStore):
        return storage_client.key_value_stores()

    if issubclass(storage_class, RequestQueue):
        return storage_client.request_queues()

    raise ValueError(f'Unknown storage class: {storage_class.__name__}')



from __future__ import annotations

from collections import OrderedDict
from collections.abc import MutableMapping
from typing import Generic, ItemsView, Iterator, TypeVar, ValuesView
from typing import OrderedDict as OrderedDictType

T = TypeVar('T')


class LRUCache(MutableMapping, Generic[T]):
    """Attempt to reimplement LRUCache from `@apify/datastructures` using `OrderedDict`."""

    def __init__(self, max_length: int) -> None:
        """Create a LRUCache with a specific max_length."""
        self._cache: OrderedDictType[str, T] = OrderedDict()
        self._max_length = max_length

    def __getitem__(self, key: str) -> T:
        """Get an item from the cache. Move it to the end if present."""
        val = self._cache[key]
        # No 'key in cache' condition since the previous line would raise KeyError
        self._cache.move_to_end(key)
        return val

    def __setitem__(self, key: str, value: T) -> None:
        """Add an item to the cache. Remove least used item if max_length exceeded."""
        # Sadly TS impl returns bool indicating whether the key was already present or not
        self._cache[key] = value
        if len(self._cache) > self._max_length:
            self._cache.popitem(last=False)

    def __delitem__(self, key: str) -> None:
        """Remove an item from the cache."""
        del self._cache[key]

    def __iter__(self) -> Iterator[str]:
        """Iterate over the keys of the cache in order of insertion."""
        return self._cache.__iter__()

    def __len__(self) -> int:
        """Get the number of items in the cache."""
        return len(self._cache)

    def values(self) -> ValuesView[T]:  # Needed so we don't mutate the cache by __getitem__
        """Iterate over the values in the cache in order of insertion."""
        return self._cache.values()

    def items(self) -> ItemsView[str, T]:  # Needed so we don't mutate the cache by __getitem__
        """Iterate over the pairs of (key, value) in the cache in order of insertion."""
        return self._cache.items()



# ruff: noqa: TCH001, TCH002, TCH003

from __future__ import annotations

from datetime import datetime
from decimal import Decimal
from enum import Enum
from typing import Annotated, Any, Generic

from pydantic import BaseModel, ConfigDict, Field
from typing_extensions import Self, TypeVar

from crawlee._utils.requests import compute_unique_key, unique_key_to_request_id
from crawlee.enqueue_strategy import EnqueueStrategy


class BaseRequestData(BaseModel):
    """Data needed to create a new crawling request."""

    model_config = ConfigDict(populate_by_name=True)

    url: Annotated[str, Field(min_length=1)]
    """URL of the web page to crawl"""

    unique_key: Annotated[str, Field(alias='uniqueKey')]
    """A unique key identifying the request. Two requests with the same `uniqueKey` are considered as pointing to the
    same URL.

    If `uniqueKey` is not provided, then it is automatically generated by normalizing the URL.
    For example, the URL of `HTTP://www.EXAMPLE.com/something/` will produce the `uniqueKey`
    of `http://www.example.com/something`.

    Pass an arbitrary non-empty text value to the `uniqueKey` property
    to override the default behavior and specify which URLs shall be considered equal.
    """

    method: str = 'get'

    payload: str | None = None

    headers: Annotated[dict[str, str] | None, Field(default_factory=dict)] = None

    user_data: Annotated[dict[str, Any], Field(alias='userData', default_factory=dict)]
    """Custom user data assigned to the request. Use this to save any request related data to the
    request's scope, keeping them accessible on retries, failures etc.
    """

    retry_count: Annotated[int, Field(alias='retryCount')] = 0

    no_retry: Annotated[bool, Field(alias='noRetry')] = False

    loaded_url: Annotated[str | None, Field(alias='loadedUrl')] = None

    handled_at: Annotated[datetime | None, Field(alias='handledAt')] = None

    @classmethod
    def from_url(
        cls,
        url: str,
        *,
        label: str | None = None,
        unique_key: str | None = None,
        **kwargs: Any,
    ) -> Self:
        """Create a new `RequestData` instance from a URL."""
        unique_key = unique_key or compute_unique_key(url)
        result = cls(url=url, unique_key=unique_key, **kwargs)

        if label is not None:
            result.user_data['label'] = label

        return result


class Request(BaseRequestData):
    """A crawling request (as returned from a request queue)."""

    id: str

    json_: str | None = None  # TODO: get rid of this
    # https://github.com/apify/crawlee-py/issues/94

    order_no: Decimal | None = None  # TODO: get rid of this
    # https://github.com/apify/crawlee-py/issues/94

    @classmethod
    def from_url(
        cls,
        url: str,
        *,
        label: str | None = None,
        unique_key: str | None = None,
        id: str | None = None,
        **kwargs: Any,
    ) -> Self:
        """Create a new `RequestData` instance from a URL."""
        unique_key = unique_key or compute_unique_key(url)
        id = id or unique_key_to_request_id(unique_key)

        result = cls(url=url, unique_key=unique_key, id=id, **kwargs)

        if label is not None:
            result.user_data['label'] = label

        return result

    @classmethod
    def from_base_request_data(cls, base_request_data: BaseRequestData, *, id: str | None = None) -> Self:
        """Create a complete Request object based on a BaseRequestData instance."""
        kwargs = base_request_data.model_dump()
        kwargs['id'] = id or unique_key_to_request_id(base_request_data.unique_key)
        return cls(**kwargs)

    @property
    def label(self) -> str | None:
        """A string used to differentiate between arbitrary request types."""
        if 'label' in self.user_data:
            return str(self.user_data['label'])
        return None

    @property
    def crawlee_data(self) -> CrawleeRequestData:
        """Crawlee-specific configuration stored in the user_data."""
        return CrawleeRequestData.model_validate(self.user_data.get('__crawlee', {}))

    @property
    def state(self) -> RequestState | None:
        """Crawlee-specific request handling state."""
        return self.crawlee_data.state

    @state.setter
    def state(self, new_state: RequestState) -> None:
        self.user_data.setdefault('__crawlee', {})
        self.user_data['__crawlee']['state'] = new_state

    @property
    def max_retries(self) -> int | None:
        """Crawlee-specific limit on the number of retries of the request."""
        return self.crawlee_data.max_retries

    @max_retries.setter
    def max_retries(self, new_max_retries: int) -> None:
        self.user_data.setdefault('__crawlee', {})
        self.user_data['__crawlee']['maxRetries'] = new_max_retries

    @property
    def session_rotation_count(self) -> int | None:
        """Crawlee-specific number of finished session rotations for the request."""
        return self.crawlee_data.session_rotation_count

    @session_rotation_count.setter
    def session_rotation_count(self, new_session_rotation_count: int) -> None:
        self.user_data.setdefault('__crawlee', {})
        self.user_data['__crawlee']['sessionRotationCount'] = new_session_rotation_count

    @property
    def enqueue_strategy(self) -> EnqueueStrategy:
        """The strategy used when enqueueing the request."""
        return (
            EnqueueStrategy(self.crawlee_data.enqueue_strategy)
            if self.crawlee_data.enqueue_strategy
            else EnqueueStrategy.ALL
        )

    @enqueue_strategy.setter
    def enqueue_strategy(self, new_enqueue_strategy: EnqueueStrategy) -> None:
        self.user_data.setdefault('__crawlee', {})
        self.user_data['__crawlee']['enqueueStrategy'] = str(new_enqueue_strategy)

    @property
    def last_proxy_tier(self) -> int | None:
        """The last proxy tier used to process the request."""
        return self.crawlee_data.last_proxy_tier

    @last_proxy_tier.setter
    def last_proxy_tier(self, new_value: int) -> None:
        self.user_data.setdefault('__crawlee', {})
        self.user_data['__crawlee']['lastProxyTier'] = new_value

    @property
    def forefront(self) -> bool:
        """Should the request be enqueued at the start of the queue?"""
        return self.crawlee_data.forefront

    @forefront.setter
    def forefront(self, new_value: bool) -> None:
        self.user_data.setdefault('__crawlee', {})
        self.user_data['__crawlee']['forefront'] = new_value


class RequestWithLock(Request):
    """A crawling request with information about locks."""

    lock_expires_at: Annotated[datetime, Field(alias='lockExpiresAt')]


class RequestState(Enum):
    """Crawlee-specific request handling state."""

    UNPROCESSED = 0
    BEFORE_NAV = 1
    AFTER_NAV = 2
    REQUEST_HANDLER = 3
    DONE = 4
    ERROR_HANDLER = 5
    ERROR = 6
    SKIPPED = 7


class CrawleeRequestData(BaseModel):
    """Crawlee-specific configuration stored in the user_data."""

    max_retries: Annotated[int | None, Field(alias='maxRetries')] = None
    """Maximum number of retries for this request. Allows to override the global `max_request_retries` option of
    `BasicCrawler`."""

    enqueue_strategy: Annotated[str | None, Field(alias='enqueueStrategy')] = None

    state: RequestState | None = None
    """Describes the request's current lifecycle state."""

    session_rotation_count: Annotated[int | None, Field(alias='sessionRotationCount')] = None

    skip_navigation: Annotated[bool, Field(alias='skipNavigation')] = False

    last_proxy_tier: Annotated[int | None, Field(alias='lastProxyTier')] = None

    forefront: Annotated[bool, Field()] = False


class BaseStorageMetadata(BaseModel):
    """Base model for storage metadata."""

    model_config = ConfigDict(populate_by_name=True)

    id: Annotated[str, Field(alias='id')]
    name: Annotated[str | None, Field(alias='name', default='')]
    accessed_at: Annotated[datetime, Field(alias='accessedAt')]
    created_at: Annotated[datetime, Field(alias='createdAt')]
    modified_at: Annotated[datetime, Field(alias='modifiedAt')]


class DatasetMetadata(BaseStorageMetadata):
    """Model for a dataset metadata."""

    model_config = ConfigDict(populate_by_name=True)

    item_count: Annotated[int, Field(alias='itemCount')]


class KeyValueStoreMetadata(BaseStorageMetadata):
    """Model for a key-value store metadata."""

    model_config = ConfigDict(populate_by_name=True)

    user_id: Annotated[str, Field(alias='userId')]


class RequestQueueMetadata(BaseStorageMetadata):
    """Model for a request queue metadata."""

    model_config = ConfigDict(populate_by_name=True)

    had_multiple_clients: Annotated[bool, Field(alias='hadMultipleClients')]
    handled_request_count: Annotated[int, Field(alias='handledRequestCount')]
    pending_request_count: Annotated[int, Field(alias='pendingRequestCount')]
    stats: Annotated[dict, Field(alias='stats')]
    total_request_count: Annotated[int, Field(alias='totalRequestCount')]
    user_id: Annotated[str, Field(alias='userId')]
    resource_directory: Annotated[str, Field(alias='resourceDirectory')]


ValueType = TypeVar('ValueType', default=Any)


class KeyValueStoreRecord(BaseModel, Generic[ValueType]):
    """Model for a key-value store record."""

    model_config = ConfigDict(populate_by_name=True)

    key: Annotated[str, Field(alias='key')]
    value: Annotated[ValueType, Field(alias='value')]
    content_type: Annotated[str | None, Field(alias='contentType', default=None)]
    filename: Annotated[str | None, Field(alias='filename', default=None)]


class KeyValueStoreRecordMetadata(BaseModel):
    """Model for a key-value store record metadata."""

    model_config = ConfigDict(populate_by_name=True)

    key: Annotated[str, Field(alias='key')]
    content_type: Annotated[str, Field(alias='contentType')]


class KeyValueStoreKeyInfo(BaseModel):
    """Model for a key-value store key info."""

    model_config = ConfigDict(populate_by_name=True)

    key: Annotated[str, Field(alias='key')]
    size: Annotated[int, Field(alias='size')]


class KeyValueStoreListKeysPage(BaseModel):
    """Model for listing keys in the key-value store."""

    model_config = ConfigDict(populate_by_name=True)

    count: Annotated[int, Field(alias='count')]
    limit: Annotated[int, Field(alias='limit')]
    is_truncated: Annotated[bool, Field(alias='isTruncated')]
    items: Annotated[list[KeyValueStoreKeyInfo], Field(alias='items', default_factory=list)]
    exclusive_start_key: Annotated[str | None, Field(alias='exclusiveStartKey', default=None)]
    next_exclusive_start_key: Annotated[str | None, Field(alias='nextExclusiveStartKey', default=None)]


class RequestQueueHeadState(BaseModel):
    """Model for the request queue head state."""

    model_config = ConfigDict(populate_by_name=True)

    was_limit_reached: Annotated[bool, Field(alias='wasLimitReached')]
    prev_limit: Annotated[int, Field(alias='prevLimit')]
    queue_modified_at: Annotated[datetime, Field(alias='queueModifiedAt')]
    query_started_at: Annotated[datetime, Field(alias='queryStartedAt')]
    had_multiple_clients: Annotated[bool, Field(alias='hadMultipleClients')]


class RequestQueueHead(BaseModel):
    """Model for the request queue head."""

    model_config = ConfigDict(populate_by_name=True)

    limit: Annotated[int | None, Field(alias='limit', default=None)]
    had_multiple_clients: Annotated[bool, Field(alias='hadMultipleClients')]
    queue_modified_at: Annotated[datetime, Field(alias='queueModifiedAt')]
    items: Annotated[list[Request], Field(alias='items', default_factory=list)]


class RequestQueueHeadWithLocks(RequestQueueHead):
    """Model for request queue head with locks."""

    lock_secs: Annotated[int, Field(alias='lockSecs')]
    items: Annotated[list[Request], Field(alias='items', default_factory=list)]


class BaseListPage(BaseModel):
    """Model for a single page of storage items returned from a collection list method.

    Args:
        count: Count of the returned objects on this page.
        offset: The offset of the first object specified in the API call.
        limit: The limit on the number of returned objects specified in the API call.
        total: Total number of objects matching the API call criteria.
        desc: Whether the listing is descending or not.
    """

    model_config = ConfigDict(populate_by_name=True)

    count: Annotated[int, Field(default=0)]
    offset: Annotated[int, Field(default=0)]
    limit: Annotated[int, Field(default=0)]
    total: Annotated[int, Field(default=0)]
    desc: Annotated[bool, Field(default=False)]


class DatasetListPage(BaseListPage):
    """Model for a single page of dataset items returned from a collection list method.

    Args:
        items: List of returned dataset items on this page.
    """

    items: Annotated[list[DatasetMetadata], Field(default_factory=list)]


class KeyValueStoreListPage(BaseListPage):
    """Model for a single page of key-value store items returned from a collection list method.

    Args:
        items: List of returned key-value store items on this page.
    """

    items: Annotated[list[KeyValueStoreMetadata], Field(default_factory=list)]


class RequestQueueListPage(BaseListPage):
    """Model for a single page of request queue items returned from a collection list method.

    Args:
        items: List of returned request queue items on this page.
    """

    items: Annotated[list[RequestQueueMetadata], Field(default_factory=list)]


class DatasetItemsListPage(BaseListPage):
    """Model for a single page of dataset items returned from a collection list method.

    Args:
        items: List of returned dataset items on this page.
    """

    items: Annotated[list[dict], Field(default_factory=list)]


class ProlongRequestLockResponse(BaseModel):
    """Response to prolong request lock calls."""

    model_config = ConfigDict(populate_by_name=True)

    lock_expires_at: Annotated[datetime, Field(alias="'lockExpiresAt'")]


class ProcessedRequest(BaseModel):
    """Represents a processed request."""

    model_config = ConfigDict(populate_by_name=True)

    id: Annotated[str, Field(alias='id')]
    unique_key: Annotated[str, Field(alias='uniqueKey')]
    was_already_present: Annotated[bool, Field(alias='wasAlreadyPresent')]
    was_already_handled: Annotated[bool, Field(alias='wasAlreadyHandled')]


class UnprocessedRequest(BaseModel):
    """Represents an unprocessed request."""

    model_config = ConfigDict(populate_by_name=True)

    unique_key: Annotated[str, Field(alias='requestUniqueKey')]
    url: Annotated[str, Field()]
    method: Annotated[str | None, Field()] = None


class BatchRequestsOperationResponse(BaseModel):
    """Response to batch request deletion calls."""

    model_config = ConfigDict(populate_by_name=True)

    processed_requests: Annotated[list[ProcessedRequest], Field(alias='processedRequests')]
    unprocessed_requests: Annotated[list[UnprocessedRequest], Field(alias='unprocessedRequests')]


class RequestListResponse(BaseModel):
    """Response to a request list call."""

    model_config = ConfigDict(populate_by_name=True)

    limit: Annotated[int, Field()]
    exclusive_start_key: Annotated[str | None, Field(alias='exclusiveStartId')]
    items: Annotated[list[Request], Field()]



# Inspiration: https://github.com/apify/crawlee/blob/v3.10.1/packages/browser-pool/src/abstract-classes/browser-plugin.ts

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Literal

if TYPE_CHECKING:
    from collections.abc import Mapping
    from types import TracebackType

    from crawlee.browsers.base_browser_controller import BaseBrowserController


class BaseBrowserPlugin(ABC):
    """An abstract base class for browser plugins.

    Browser plugins act as wrappers around browser automation tools like Playwright,
    providing a unified interface for interacting with browsers.
    """

    AUTOMATION_LIBRARY: str | None = None
    """The name of the automation library that the plugin is managing."""

    @property
    @abstractmethod
    def browser_type(self) -> Literal['chromium', 'firefox', 'webkit']:
        """Return the browser type name."""

    @property
    @abstractmethod
    def browser_options(self) -> Mapping[str, Any]:
        """Return the options for a new browser."""

    @property
    @abstractmethod
    def page_options(self) -> Mapping[str, Any]:
        """Return the options for a new page."""

    @property
    @abstractmethod
    def max_open_pages_per_browser(self) -> int:
        """Return the maximum number of pages that can be opened in a single browser."""

    @abstractmethod
    async def __aenter__(self) -> BaseBrowserPlugin:
        """Enter the context manager and initialize the browser plugin."""

    @abstractmethod
    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        exc_traceback: TracebackType | None,
    ) -> None:
        """Exit the context manager and close the browser plugin."""

    @abstractmethod
    async def new_browser(self) -> BaseBrowserController:
        """Create a new browser instance.

        Returns:
            A new browser instance wrapped in a controller.
        """



from __future__ import annotations

import json
import mimetypes
import os
import pathlib
from datetime import datetime, timezone
from decimal import Decimal
from logging import getLogger
from typing import TYPE_CHECKING

import aiofiles
from aiofiles.os import makedirs

from crawlee._utils.data_processing import maybe_parse_body
from crawlee._utils.file import json_dumps
from crawlee.consts import METADATA_FILENAME
from crawlee.models import (
    DatasetMetadata,
    KeyValueStoreMetadata,
    KeyValueStoreRecord,
    KeyValueStoreRecordMetadata,
    Request,
    RequestQueueMetadata,
)
from crawlee.storages.dataset import Dataset
from crawlee.storages.key_value_store import KeyValueStore
from crawlee.storages.request_queue import RequestQueue

if TYPE_CHECKING:
    from crawlee.memory_storage_client.dataset_client import DatasetClient
    from crawlee.memory_storage_client.key_value_store_client import KeyValueStoreClient
    from crawlee.memory_storage_client.memory_storage_client import MemoryStorageClient, TResourceClient
    from crawlee.memory_storage_client.request_queue_client import RequestQueueClient

logger = getLogger(__name__)


async def persist_metadata_if_enabled(*, data: dict, entity_directory: str, write_metadata: bool) -> None:
    """Updates or writes metadata to a specified directory.

    The function writes a given metadata dictionary to a JSON file within a specified directory.
    The writing process is skipped if `write_metadata` is False. Before writing, it ensures that
    the target directory exists, creating it if necessary.

    Args:
        data: A dictionary containing metadata to be written.
        entity_directory: The directory path where the metadata file should be stored.
        write_metadata: A boolean flag indicating whether the metadata should be written to file.
    """
    # Skip metadata write; ensure directory exists first
    if not write_metadata:
        return

    # Ensure the directory for the entity exists
    await makedirs(entity_directory, exist_ok=True)

    # Write the metadata to the file
    file_path = os.path.join(entity_directory, METADATA_FILENAME)
    async with aiofiles.open(file_path, mode='wb') as f:
        s = await json_dumps(data)
        await f.write(s.encode('utf-8'))


def find_or_create_client_by_id_or_name_inner(
    resource_client_class: type[TResourceClient],
    memory_storage_client: MemoryStorageClient,
    id: str | None = None,
    name: str | None = None,
) -> TResourceClient | None:
    """Locates or creates a new storage client based on the given ID or name.

    This method attempts to find a storage client in the memory cache first. If not found,
    it tries to locate a storage directory by name. If still not found, it searches through
    storage directories for a matching ID or name in their metadata. If none exists, and the
    specified ID is 'default', it checks for a default storage directory. If a storage client
    is found or created, it is added to the memory cache. If no storage client can be located or
    created, the method returns None.

    Args:
        resource_client_class: The class of the resource client.
        memory_storage_client: The memory storage client used to store and retrieve storage clients.
        id: The unique identifier for the storage client. Defaults to None.
        name: The name of the storage client. Defaults to None.

    Raises:
        ValueError: If both id and name are None.

    Returns:
        The found or created storage client, or None if no client could be found or created.
    """
    from crawlee.memory_storage_client.dataset_client import DatasetClient
    from crawlee.memory_storage_client.key_value_store_client import KeyValueStoreClient
    from crawlee.memory_storage_client.request_queue_client import RequestQueueClient

    if id is None and name is None:
        raise ValueError('Either id or name must be specified.')

    # First check memory cache
    found = memory_storage_client.get_cached_resource_client(resource_client_class, id, name)

    if found is not None:
        return found

    storage_path = _determine_storage_path(resource_client_class, memory_storage_client, id, name)

    if not storage_path:
        return None

    # Create from directory if storage path is found
    if issubclass(resource_client_class, DatasetClient):
        resource_client = create_dataset_from_directory(storage_path, memory_storage_client, id, name)
    elif issubclass(resource_client_class, KeyValueStoreClient):
        resource_client = create_kvs_from_directory(storage_path, memory_storage_client, id, name)
    elif issubclass(resource_client_class, RequestQueueClient):
        resource_client = create_rq_from_directory(storage_path, memory_storage_client, id, name)
    else:
        raise TypeError('Invalid resource client class.')

    memory_storage_client.add_resource_client_to_cache(resource_client)
    return resource_client  # pyright: ignore


async def get_or_create_inner(
    *,
    memory_storage_client: MemoryStorageClient,
    storage_client_cache: list[TResourceClient],
    resource_client_class: type[TResourceClient],
    name: str | None = None,
    id: str | None = None,
) -> TResourceClient:
    """Retrieve a named storage, or create a new one when it doesn't exist.

    Args:
        memory_storage_client: The memory storage client.
        storage_client_cache: The cache of storage clients.
        resource_client_class: The class of the storage to retrieve or create.
        name: The name of the storage to retrieve or create.
        id: ID of the storage to retrieve or create.

    Returns:
        The retrieved or newly-created storage.
    """
    # If the name or id is provided, try to find the dataset in the cache
    if name or id:
        found = find_or_create_client_by_id_or_name_inner(
            resource_client_class=resource_client_class,
            memory_storage_client=memory_storage_client,
            name=name,
            id=id,
        )
        if found:
            return found

    # Otherwise, create a new one and add it to the cache
    resource_client = resource_client_class(
        id=id,
        name=name,
        memory_storage_client=memory_storage_client,
    )

    storage_client_cache.append(resource_client)

    # Write to the disk
    await persist_metadata_if_enabled(
        data=resource_client.resource_info.model_dump(),
        entity_directory=resource_client.resource_directory,
        write_metadata=memory_storage_client.write_metadata,
    )

    return resource_client


def create_dataset_from_directory(
    storage_directory: str,
    memory_storage_client: MemoryStorageClient,
    id: str | None = None,
    name: str | None = None,
) -> DatasetClient:
    from crawlee.memory_storage_client.dataset_client import DatasetClient

    item_count = 0
    created_at = datetime.now(timezone.utc)
    accessed_at = datetime.now(timezone.utc)
    modified_at = datetime.now(timezone.utc)

    # Load metadata if it exists
    metadata_filepath = os.path.join(storage_directory, METADATA_FILENAME)

    if os.path.exists(metadata_filepath):
        with open(metadata_filepath, encoding='utf-8') as f:
            json_content = json.load(f)
            resource_info = DatasetMetadata(**json_content)

        id = resource_info.id
        name = resource_info.name
        item_count = resource_info.item_count
        created_at = resource_info.created_at
        accessed_at = resource_info.accessed_at
        modified_at = resource_info.modified_at

    # Load dataset entries
    entries: dict[str, dict] = {}
    has_seen_metadata_file = False

    for entry in os.scandir(storage_directory):
        if entry.is_file():
            if entry.name == METADATA_FILENAME:
                has_seen_metadata_file = True
                continue

            with open(os.path.join(storage_directory, entry.name), encoding='utf-8') as f:
                entry_content = json.load(f)

            entry_name = entry.name.split('.')[0]
            entries[entry_name] = entry_content

            if not has_seen_metadata_file:
                item_count += 1

    # Create new dataset client
    new_client = DatasetClient(
        memory_storage_client=memory_storage_client,
        id=id,
        name=name,
        created_at=created_at,
        accessed_at=accessed_at,
        modified_at=modified_at,
        item_count=item_count,
    )

    new_client.dataset_entries.update(entries)
    return new_client


def create_kvs_from_directory(
    storage_directory: str,
    memory_storage_client: MemoryStorageClient,
    id: str | None = None,
    name: str | None = None,
) -> KeyValueStoreClient:
    from crawlee.memory_storage_client.key_value_store_client import KeyValueStoreClient

    created_at = datetime.now(timezone.utc)
    accessed_at = datetime.now(timezone.utc)
    modified_at = datetime.now(timezone.utc)

    # Load metadata if it exists
    metadata_filepath = os.path.join(storage_directory, METADATA_FILENAME)

    if os.path.exists(metadata_filepath):
        with open(metadata_filepath, encoding='utf-8') as f:
            json_content = json.load(f)
            resource_info = KeyValueStoreMetadata(**json_content)

        id = resource_info.id
        name = resource_info.name
        created_at = resource_info.created_at
        accessed_at = resource_info.accessed_at
        modified_at = resource_info.modified_at

    # Create new KVS client
    new_client = KeyValueStoreClient(
        memory_storage_client=memory_storage_client,
        id=id,
        name=name,
        accessed_at=accessed_at,
        created_at=created_at,
        modified_at=modified_at,
    )

    # Scan the KVS folder, check each entry in there and parse it as a store record
    for entry in os.scandir(storage_directory):
        if not entry.is_file():
            continue

        # Ignore metadata files on their own
        if entry.name.endswith(METADATA_FILENAME):
            continue

        # Try checking if this file has a metadata file associated with it
        record_metadata = None
        record_metadata_filepath = os.path.join(storage_directory, f'{entry.name}.__metadata__.json')

        if os.path.exists(record_metadata_filepath):
            with open(record_metadata_filepath, encoding='utf-8') as metadata_file:
                try:
                    json_content = json.load(metadata_file)
                    record_metadata = KeyValueStoreRecordMetadata(**json_content)

                except Exception:
                    logger.warning(
                        f'Metadata of key-value store entry "{entry.name}" for store {name or id} could '
                        'not be parsed. The metadata file will be ignored.',
                        exc_info=True,
                    )

        if not record_metadata:
            content_type, _ = mimetypes.guess_type(entry.name)
            if content_type is None:
                content_type = 'application/octet-stream'

            record_metadata = KeyValueStoreRecordMetadata(
                key=pathlib.Path(entry.name).stem,
                content_type=content_type,
            )

        with open(os.path.join(storage_directory, entry.name), 'rb') as f:
            file_content = f.read()

        try:
            maybe_parse_body(file_content, record_metadata.content_type)
        except Exception:
            record_metadata.content_type = 'application/octet-stream'
            logger.warning(
                f'Key-value store entry "{record_metadata.key}" for store {name or id} could not be parsed.'
                'The entry will be assumed as binary.',
                exc_info=True,
            )

        new_client.records[record_metadata.key] = KeyValueStoreRecord(
            key=record_metadata.key,
            content_type=record_metadata.content_type,
            filename=entry.name,
            value=file_content,
        )

    return new_client


def create_rq_from_directory(
    storage_directory: str,
    memory_storage_client: MemoryStorageClient,
    id: str | None = None,
    name: str | None = None,
) -> RequestQueueClient:
    from crawlee.memory_storage_client.request_queue_client import RequestQueueClient

    created_at = datetime.now(timezone.utc)
    accessed_at = datetime.now(timezone.utc)
    modified_at = datetime.now(timezone.utc)
    handled_request_count = 0
    pending_request_count = 0

    # Load metadata if it exists
    metadata_filepath = os.path.join(storage_directory, METADATA_FILENAME)

    if os.path.exists(metadata_filepath):
        with open(metadata_filepath, encoding='utf-8') as f:
            json_content = json.load(f)
            resource_info = RequestQueueMetadata(**json_content)

        id = resource_info.id
        name = resource_info.name
        created_at = resource_info.created_at
        accessed_at = resource_info.accessed_at
        modified_at = resource_info.modified_at
        handled_request_count = resource_info.handled_request_count
        pending_request_count = resource_info.pending_request_count

    # Load request entries
    entries: dict[str, Request] = {}

    for entry in os.scandir(storage_directory):
        if entry.is_file():
            if entry.name == METADATA_FILENAME:
                continue

            with open(os.path.join(storage_directory, entry.name), encoding='utf-8') as f:
                content = json.load(f)

            request = Request(**content)
            order_no = request.order_no
            if order_no:
                request.order_no = Decimal(order_no)

            entries[request.id] = request

    # Create new RQ client
    new_client = RequestQueueClient(
        memory_storage_client=memory_storage_client,
        id=id,
        name=name,
        accessed_at=accessed_at,
        created_at=created_at,
        modified_at=modified_at,
        handled_request_count=handled_request_count,
        pending_request_count=pending_request_count,
    )

    new_client.requests.update(entries)
    return new_client


def _determine_storage_path(
    resource_client_class: type[TResourceClient],
    memory_storage_client: MemoryStorageClient,
    id: str | None = None,
    name: str | None = None,
) -> str | None:
    from crawlee.memory_storage_client.dataset_client import DatasetClient
    from crawlee.memory_storage_client.key_value_store_client import KeyValueStoreClient
    from crawlee.memory_storage_client.request_queue_client import RequestQueueClient
    from crawlee.storages._creation_management import _get_default_storage_id

    configuration = memory_storage_client._configuration  # noqa: SLF001

    if issubclass(resource_client_class, DatasetClient):
        storages_dir = memory_storage_client.datasets_directory
        default_id = _get_default_storage_id(configuration, Dataset)
    elif issubclass(resource_client_class, KeyValueStoreClient):
        storages_dir = memory_storage_client.key_value_stores_directory
        default_id = _get_default_storage_id(configuration, KeyValueStore)
    elif issubclass(resource_client_class, RequestQueueClient):
        storages_dir = memory_storage_client.request_queues_directory
        default_id = _get_default_storage_id(configuration, RequestQueue)
    else:
        raise TypeError('Invalid resource client class.')

    # Try to find by name directly from directories
    if name:
        possible_storage_path = os.path.join(storages_dir, name)
        if os.access(possible_storage_path, os.F_OK):
            return possible_storage_path

    # If not found, try finding by metadata
    if os.access(storages_dir, os.F_OK):
        for entry in os.scandir(storages_dir):
            if entry.is_dir():
                metadata_path = os.path.join(entry.path, METADATA_FILENAME)
                if os.access(metadata_path, os.F_OK):
                    with open(metadata_path, encoding='utf-8') as metadata_file:
                        metadata = json.load(metadata_file)
                    if (id and metadata.get('id') == id) or (name and metadata.get('name') == name):
                        return entry.path

    # Check for default storage directory as a last resort
    if id == default_id:
        possible_storage_path = os.path.join(storages_dir, default_id)
        if os.access(possible_storage_path, os.F_OK):
            return possible_storage_path

    return None



# ruff: noqa: TCH003
from __future__ import annotations

import os
from contextlib import suppress
from datetime import datetime, timezone
from logging import getLogger
from typing import Annotated

import psutil
from pydantic import BaseModel, ConfigDict, Field, PlainSerializer, PlainValidator

from crawlee._utils.byte_size import ByteSize

logger = getLogger(__name__)


class CpuInfo(BaseModel):
    """Information about the CPU usage.

    Args:
        used_ratio: The ratio of CPU currently in use, represented as a float between 0 and 1.
        created_at: The time at which the measurement was taken.
    """

    model_config = ConfigDict(populate_by_name=True)

    used_ratio: Annotated[float, Field(alias='usedRatio')]
    created_at: datetime = Field(
        alias='createdAt',
        default_factory=lambda: datetime.now(timezone.utc),
    )


class MemoryInfo(BaseModel):
    """Information about the memory usage.

    Args:
        total_size: Total memory available in the system.
        current_size: Memory usage of the current Python process and its children.
        created_at: The time at which the measurement was taken.
    """

    model_config = ConfigDict(populate_by_name=True)

    total_size: Annotated[
        ByteSize, PlainValidator(ByteSize.validate), PlainSerializer(lambda size: size.bytes), Field(alias='totalSize')
    ]
    current_size: Annotated[
        ByteSize,
        PlainValidator(ByteSize.validate),
        PlainSerializer(lambda size: size.bytes),
        Field(alias='currentSize'),
    ]
    created_at: datetime = Field(
        alias='createdAt',
        default_factory=lambda: datetime.now(timezone.utc),
    )


def get_cpu_info() -> CpuInfo:
    """Retrieves the current CPU usage.

    It utilizes the `psutil` library. Function `psutil.cpu_percent()` returns a float representing the current
    system-wide CPU utilization as a percentage.
    """
    logger.debug('Calling get_cpu_info()...')
    cpu_percent = psutil.cpu_percent(interval=0.1)
    return CpuInfo(used_ratio=cpu_percent / 100)


def get_memory_info() -> MemoryInfo:
    """Retrieves the current memory usage of the process and its children.

    It utilizes the `psutil` library.
    """
    logger.debug('Calling get_memory_info()...')
    current_process = psutil.Process(os.getpid())

    # Retrieve the Resident Set Size (RSS) of the current process. RSS is the portion of memory
    # occupied by a process that is held in RAM.
    current_size_bytes = int(current_process.memory_info().rss)

    for child in current_process.children(recursive=True):
        # Ignore any NoSuchProcess exception that might occur if a child process ends before we retrieve
        # its memory usage.
        with suppress(psutil.NoSuchProcess):
            current_size_bytes += int(child.memory_info().rss)

    total_size_bytes = psutil.virtual_memory().total

    return MemoryInfo(
        total_size=ByteSize(total_size_bytes),
        current_size=ByteSize(current_size_bytes),
    )



from __future__ import annotations

import asyncio
import logging
from typing import TYPE_CHECKING, Any, AsyncGenerator, Iterable, Literal

from bs4 import BeautifulSoup, Tag
from typing_extensions import Unpack

from crawlee._utils.blocked import RETRY_CSS_SELECTORS
from crawlee.basic_crawler import BasicCrawler, BasicCrawlerOptions, ContextPipeline
from crawlee.basic_crawler.errors import SessionError
from crawlee.beautifulsoup_crawler.types import BeautifulSoupCrawlingContext
from crawlee.enqueue_strategy import EnqueueStrategy
from crawlee.http_clients import HttpxClient
from crawlee.http_crawler import HttpCrawlingContext
from crawlee.models import BaseRequestData

if TYPE_CHECKING:
    from crawlee.basic_crawler.types import AddRequestsKwargs, BasicCrawlingContext


class BeautifulSoupCrawler(BasicCrawler[BeautifulSoupCrawlingContext]):
    """A crawler that fetches the request URL using `httpx` and parses the result with `BeautifulSoup`."""

    def __init__(
        self,
        *,
        parser: Literal['html.parser', 'lxml', 'xml', 'html5lib'] = 'lxml',
        additional_http_error_status_codes: Iterable[int] = (),
        ignore_http_error_status_codes: Iterable[int] = (),
        **kwargs: Unpack[BasicCrawlerOptions[BeautifulSoupCrawlingContext]],
    ) -> None:
        """Initialize the BeautifulSoupCrawler.

        Args:
            parser: The type of parser that should be used by BeautifulSoup

            additional_http_error_status_codes: HTTP status codes that should be considered errors (and trigger a retry)

            ignore_http_error_status_codes: HTTP status codes that are normally considered errors but we want to treat
                them as successful

            kwargs: Arguments to be forwarded to the underlying BasicCrawler
        """
        self._parser = parser

        kwargs['_context_pipeline'] = (
            ContextPipeline()
            .compose(self._make_http_request)
            .compose(self._parse_http_response)
            .compose(self._handle_blocked_request)
        )

        kwargs.setdefault(
            'http_client',
            HttpxClient(
                additional_http_error_status_codes=additional_http_error_status_codes,
                ignore_http_error_status_codes=ignore_http_error_status_codes,
            ),
        )

        kwargs.setdefault('_logger', logging.getLogger(__name__))

        super().__init__(**kwargs)

    async def _make_http_request(self, context: BasicCrawlingContext) -> AsyncGenerator[HttpCrawlingContext, None]:
        result = await self._http_client.crawl(
            context.request,
            context.session,
            context.proxy_info,
            self._statistics,
        )

        yield HttpCrawlingContext(
            request=context.request,
            session=context.session,
            proxy_info=context.proxy_info,
            add_requests=context.add_requests,
            send_request=context.send_request,
            push_data=context.push_data,
            log=context.log,
            http_response=result.http_response,
        )

    async def _handle_blocked_request(
        self, crawling_context: BeautifulSoupCrawlingContext
    ) -> AsyncGenerator[BeautifulSoupCrawlingContext, None]:
        if self._retry_on_blocked:
            status_code = crawling_context.http_response.status_code

            if crawling_context.session and crawling_context.session.is_blocked_status_code(status_code=status_code):
                raise SessionError(f'Assuming the session is blocked based on HTTP status code {status_code}')

            matched_selectors = [
                selector for selector in RETRY_CSS_SELECTORS if crawling_context.soup.select_one(selector) is not None
            ]

            if matched_selectors:
                raise SessionError(
                    'Assuming the session is blocked - '
                    f"HTTP response matched the following selectors: {'; '.join(matched_selectors)}"
                )

        yield crawling_context

    async def _parse_http_response(
        self,
        context: HttpCrawlingContext,
    ) -> AsyncGenerator[BeautifulSoupCrawlingContext, None]:
        soup = await asyncio.to_thread(lambda: BeautifulSoup(context.http_response.read(), self._parser))

        async def enqueue_links(
            *,
            selector: str = 'a',
            label: str | None = None,
            user_data: dict[str, Any] | None = None,
            **kwargs: Unpack[AddRequestsKwargs],
        ) -> None:
            kwargs.setdefault('strategy', EnqueueStrategy.SAME_HOSTNAME)

            requests = list[BaseRequestData]()
            user_data = user_data or {}

            link: Tag
            for link in soup.select(selector):
                link_user_data = user_data

                if label is not None:
                    link_user_data.setdefault('label', label)

                if (href := link.attrs.get('href')) is not None:
                    requests.append(BaseRequestData.from_url(href, user_data=link_user_data))

            await context.add_requests(requests, **kwargs)

        yield BeautifulSoupCrawlingContext(
            request=context.request,
            session=context.session,
            proxy_info=context.proxy_info,
            enqueue_links=enqueue_links,
            add_requests=context.add_requests,
            send_request=context.send_request,
            push_data=context.push_data,
            log=context.log,
            http_response=context.http_response,
            soup=soup,
        )



# Inspiration: https://github.com/apify/crawlee/tree/v3.10.1/packages/browser-pool/

from __future__ import annotations

import asyncio
import itertools
from collections import defaultdict
from datetime import timedelta
from logging import getLogger
from typing import TYPE_CHECKING, Any, Literal
from weakref import WeakValueDictionary

from crawlee._utils.crypto import crypto_random_object_id
from crawlee._utils.recurring_task import RecurringTask
from crawlee.browsers.base_browser_controller import BaseBrowserController
from crawlee.browsers.playwright_browser_plugin import PlaywrightBrowserPlugin
from crawlee.browsers.types import CrawleePage

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence
    from types import TracebackType

    from crawlee.browsers.base_browser_plugin import BaseBrowserPlugin
    from crawlee.proxy_configuration import ProxyInfo

logger = getLogger(__name__)


class BrowserPool:
    """Manages a pool of browsers and their pages, handling lifecycle events and resource allocation.

    This class is responsible for opening and closing browsers, managing pages within those browsers,
    and handling the overall lifecycle of these resources. It provides flexible configuration via
    constructor options, which include various hooks that allow for the insertion of custom behavior
    at different stages of the browser and page lifecycles.

    The browsers in the pool can be in one of three states: active, inactive, or closed.
    """

    _GENERATED_PAGE_ID_LENGTH = 8
    """The length of the newly generated page ID."""

    def __init__(
        self,
        plugins: Sequence[BaseBrowserPlugin] | None = None,
        *,
        operation_timeout: timedelta = timedelta(seconds=15),
        browser_inactive_threshold: timedelta = timedelta(seconds=10),
        identify_inactive_browsers_interval: timedelta = timedelta(seconds=20),
        close_inactive_browsers_interval: timedelta = timedelta(seconds=30),
    ) -> None:
        """Create a new instance.

        Args:
            plugins: Browser plugins serve as wrappers around various browser automation libraries,
                providing a consistent interface across different libraries.
            operation_timeout: Operations of the underlying automation libraries, such as launching a browser
                or opening a new page, can sometimes get stuck. To prevent `BrowserPool` from becoming unresponsive,
                we add a timeout to these operations.
            browser_inactive_threshold: The period of inactivity after which a browser is considered as inactive.
            identify_inactive_browsers_interval: The period of inactivity after which a browser is considered
                as retired.
            close_inactive_browsers_interval: The interval at which the pool checks for inactive browsers
                and closes them. The browser is considered as inactive if it has no active pages and has been idle
                for the specified period.
        """
        self._plugins = plugins or [PlaywrightBrowserPlugin()]
        self._operation_timeout = operation_timeout
        self._browser_inactive_threshold = browser_inactive_threshold

        self._active_browsers = list[BaseBrowserController]()
        """A list of browsers currently active and being used to open pages."""

        self._inactive_browsers = list[BaseBrowserController]()
        """A list of browsers currently inactive and not being used to open new pages,
        but may still contain open pages."""

        self._identify_inactive_browsers_task = RecurringTask(
            self._identify_inactive_browsers,
            identify_inactive_browsers_interval,
        )

        self._close_inactive_browsers_task = RecurringTask(
            self._close_inactive_browsers,
            close_inactive_browsers_interval,
        )

        self._total_pages_count = 0
        self._pages = WeakValueDictionary[str, CrawleePage]()  # Track the pages in the pool
        self._plugins_cycle = itertools.cycle(self._plugins)  # Cycle through the plugins

    @classmethod
    def with_default_plugin(
        cls,
        *,
        headless: bool | None = None,
        browser_type: Literal['chromium', 'firefox', 'webkit'] | None = None,
        **kwargs: Any,
    ) -> BrowserPool:
        """Create a new instance with a single `BaseBrowserPlugin` configured with the provided options.

        Args:
            headless: Whether to run the browser in headless mode.
            browser_type: The type of browser to launch ('chromium', 'firefox', or 'webkit').
            kwargs: Additional arguments for default constructor.
        """
        plugin_options: dict = defaultdict(dict)

        if headless is not None:
            plugin_options['browser_options']['headless'] = headless

        if browser_type:
            plugin_options['browser_type'] = browser_type

        plugin = PlaywrightBrowserPlugin(**plugin_options)
        return cls(plugins=[plugin], **kwargs)

    @property
    def plugins(self) -> Sequence[BaseBrowserPlugin]:
        """Return the browser plugins."""
        return self._plugins

    @property
    def active_browsers(self) -> Sequence[BaseBrowserController]:
        """Return the active browsers in the pool."""
        return self._active_browsers

    @property
    def inactive_browsers(self) -> Sequence[BaseBrowserController]:
        """Return the inactive browsers in the pool."""
        return self._inactive_browsers

    @property
    def pages(self) -> Mapping[str, CrawleePage]:
        """Return the pages in the pool."""
        return self._pages

    @property
    def total_pages_count(self) -> int:
        """Returns the total number of pages opened since the browser pool was launched."""
        return self._total_pages_count

    async def __aenter__(self) -> BrowserPool:
        """Enter the context manager and initialize all browser plugins."""
        logger.debug('Initializing browser pool.')

        # Start the recurring tasks for identifying and closing inactive browsers
        self._identify_inactive_browsers_task.start()
        self._close_inactive_browsers_task.start()

        timeout = self._operation_timeout.total_seconds()

        try:
            for plugin in self._plugins:
                await asyncio.wait_for(plugin.__aenter__(), timeout)
        except asyncio.TimeoutError:
            logger.warning(f'Initializing of the browser plugin {plugin} timed out, will be skipped.')

        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        exc_traceback: TracebackType | None,
    ) -> None:
        """Exit the context manager and close all browser plugins."""
        logger.debug('Closing browser pool.')

        await self._identify_inactive_browsers_task.stop()
        await self._close_inactive_browsers_task.stop()

        for browser in self._active_browsers + self._inactive_browsers:
            await browser.close(force=True)

        for plugin in self._plugins:
            await plugin.__aexit__(exc_type, exc_value, exc_traceback)

    async def new_page(
        self,
        *,
        page_id: str | None = None,
        browser_plugin: BaseBrowserPlugin | None = None,
        proxy_info: ProxyInfo | None = None,
    ) -> CrawleePage:
        """Opens a new page in a browser using the specified or a random browser plugin.

        Args:
            page_id: The ID to assign to the new page. If not provided, a random ID is generated.
            browser_plugin: browser_plugin: The browser plugin to use for creating the new page.
                If not provided, the next plugin in the rotation is used.
            proxy_info: The proxy configuration to use for the new page.

        Returns:
            The newly created browser page.
        """
        if page_id in self.pages:
            raise ValueError(f'Page with ID: {page_id} already exists.')

        if browser_plugin and browser_plugin not in self.plugins:
            raise ValueError('Provided browser_plugin is not one of the plugins used by BrowserPool.')

        page_id = page_id or crypto_random_object_id(self._GENERATED_PAGE_ID_LENGTH)
        plugin = browser_plugin or next(self._plugins_cycle)

        return await self._get_new_page(page_id, plugin, proxy_info)

    async def new_page_with_each_plugin(self) -> Sequence[CrawleePage]:
        """Create a new page with each browser plugin in the pool.

        This method is useful for running scripts in multiple environments simultaneously, typically for testing
        or website analysis. Each page is created using a different browser plugin, allowing you to interact
        with various browser types concurrently.

        Returns:
            A list of newly created pages, one for each plugin in the pool.
        """
        pages_coroutines = [self.new_page(browser_plugin=plugin) for plugin in self._plugins]
        return await asyncio.gather(*pages_coroutines)

    async def _get_new_page(
        self,
        page_id: str,
        plugin: BaseBrowserPlugin,
        proxy_info: ProxyInfo | None,
    ) -> CrawleePage:
        """Internal method to initialize a new page in a browser using the specified plugin."""
        timeout = self._operation_timeout.total_seconds()
        browser = self._pick_browser_with_free_capacity(plugin)

        try:
            if not browser:
                browser = await asyncio.wait_for(self._launch_new_browser(plugin), timeout)
            page = await asyncio.wait_for(
                browser.new_page(page_options=plugin.page_options, proxy_info=proxy_info),
                timeout,
            )
        except asyncio.TimeoutError as exc:
            raise TimeoutError(f'Creating a new page with plugin {plugin} timed out.') from exc
        except RuntimeError as exc:
            raise RuntimeError('Browser pool is not initialized.') from exc

        crawlee_page = CrawleePage(id=page_id, page=page, browser_type=plugin.browser_type)
        self._pages[page_id] = crawlee_page
        self._total_pages_count += 1
        return crawlee_page

    def _pick_browser_with_free_capacity(
        self,
        browser_plugin: BaseBrowserPlugin,
    ) -> BaseBrowserController | None:
        """Pick a browser with free capacity that matches the specified plugin."""
        for browser in self._active_browsers:
            if browser.has_free_capacity and browser.AUTOMATION_LIBRARY == browser_plugin.AUTOMATION_LIBRARY:
                return browser

        return None

    async def _launch_new_browser(self, plugin: BaseBrowserPlugin) -> BaseBrowserController:
        """Launch a new browser instance using the specified plugin."""
        browser = await plugin.new_browser()
        self._active_browsers.append(browser)
        return browser

    def _identify_inactive_browsers(self) -> None:
        """Identify inactive browsers and move them to the inactive list if their idle time exceeds the threshold."""
        for browser in self._active_browsers:
            if browser.idle_time >= self._browser_inactive_threshold:
                self._active_browsers.remove(browser)
                self._inactive_browsers.append(browser)

    async def _close_inactive_browsers(self) -> None:
        """Close the browsers that have no active pages and have been idle for a certain period."""
        for browser in self._inactive_browsers:
            if not browser.pages:
                await browser.close()
                self._inactive_browsers.remove(browser)



from __future__ import annotations

import asyncio
from logging import getLogger
from typing import TYPE_CHECKING, Callable

if TYPE_CHECKING:
    from datetime import timedelta

logger = getLogger(__name__)


class RecurringTask:
    """Class for creating and managing recurring tasks.

    Attributes:
        func: The function to be executed repeatedly.
        delay: The time delay (in seconds) between function calls.
        task: The underlying task object.
    """

    def __init__(self, func: Callable, delay: timedelta) -> None:
        logger.debug(f'Calling RecurringTask.__init__(func={func.__name__}, delay={delay})...')
        self.func = func
        self.delay = delay
        self.task: asyncio.Task | None = None

    async def _wrapper(self) -> None:
        """Internal method that repeatedly executes the provided function with the specified delay."""
        sleep_time_secs = self.delay.total_seconds()
        while True:
            await self.func() if asyncio.iscoroutinefunction(self.func) else self.func()
            await asyncio.sleep(sleep_time_secs)

    def start(self) -> None:
        """Start the recurring task execution."""
        self.task = asyncio.create_task(self._wrapper(), name=f'Task-recurring-{self.func.__name__}')

    async def stop(self) -> None:
        """Stop the recurring task execution."""
        if self.task:
            self.task.cancel()
            # Ensure the task has a chance to properly handle the cancellation and any potential exceptions.
            await asyncio.gather(self.task, return_exceptions=True)



from __future__ import annotations

from typing import Generic

from typing_extensions import TypeVar

from crawlee.basic_crawler.types import BasicCrawlingContext


class UserDefinedErrorHandlerError(Exception):
    """Wraps an exception thrown from an user-defined error handler."""


TCrawlingContext = TypeVar('TCrawlingContext', bound=BasicCrawlingContext, default=BasicCrawlingContext)


class SessionError(Exception):
    """Errors of `SessionError` type will trigger a session rotation.

    This error doesn't respect the `max_request_retries` option and has a separate limit of `max_session_rotations`.
    """


class ProxyError(SessionError):
    """Raised when a proxy is being blocked or malfunctions."""


class RequestHandlerError(Exception, Generic[TCrawlingContext]):
    """Wraps an exception thrown from a request handler (router) and extends it with crawling context."""

    def __init__(self, wrapped_exception: Exception, crawling_context: TCrawlingContext) -> None:
        self.wrapped_exception = wrapped_exception
        self.crawling_context = crawling_context


class ContextPipelineInitializationError(Exception):
    """Wraps an exception thrown in the initialization step of a context pipeline middleware.

    We may not have the complete context at this point, so only `BasicCrawlingContext` is provided.
    """

    def __init__(self, wrapped_exception: Exception, crawling_context: BasicCrawlingContext) -> None:
        self.wrapped_exception = wrapped_exception
        self.crawling_context = crawling_context


class ContextPipelineFinalizationError(Exception):
    """Wraps an exception thrown in the finalization step of a context pipeline middleware.

    We may not have the complete context at this point, so only `BasicCrawlingContext` is provided.
    """

    def __init__(self, wrapped_exception: Exception, crawling_context: BasicCrawlingContext) -> None:
        self.wrapped_exception = wrapped_exception
        self.crawling_context = crawling_context


class ContextPipelineInterruptedError(Exception):
    """May be thrown in the initialization phase of a middleware to signal that the request should not be processed."""



from __future__ import annotations

import asyncio
from collections import OrderedDict
from datetime import datetime, timedelta, timezone
from logging import getLogger
from typing import TYPE_CHECKING
from typing import OrderedDict as OrderedDictType

from typing_extensions import override

from crawlee._utils.crypto import crypto_random_object_id
from crawlee._utils.lru_cache import LRUCache
from crawlee._utils.requests import unique_key_to_request_id
from crawlee._utils.wait import wait_for_all_tasks_for_finish
from crawlee.models import (
    BaseRequestData,
    ProcessedRequest,
    Request,
    RequestQueueHeadState,
    RequestQueueMetadata,
)
from crawlee.storages.base_storage import BaseStorage
from crawlee.storages.request_provider import RequestProvider

if TYPE_CHECKING:
    from collections.abc import Sequence

    from crawlee.base_storage_client import BaseStorageClient
    from crawlee.configuration import Configuration

logger = getLogger(__name__)


class RequestQueue(BaseStorage, RequestProvider):
    """Represents a queue storage for HTTP requests to crawl.

    Manages a queue of requests with unique URLs for structured deep web crawling with support for both breadth-first
    and depth-first orders. This queue is designed for crawling websites by starting with initial URLs and recursively
    following links. Each URL is uniquely identified by a `unique_key` field, which can be overridden to add the same
    URL multiple times under different keys.

    Local storage path (if `CRAWLEE_STORAGE_DIR` is set):
    `{CRAWLEE_STORAGE_DIR}/request_queues/{QUEUE_ID}/{REQUEST_ID}.json`, where `{QUEUE_ID}` is the request
    queue's ID (default or specified) and `{REQUEST_ID}` is the request's ID.

    Usage includes creating or opening existing queues by ID or name, with named queues retained indefinitely and
    unnamed queues expiring after 7 days unless specified otherwise. Supports mutable operations—URLs can be added
    and deleted.

    Usage:
        rq = await RequestQueue.open(id='my_rq_id')
    """

    _API_PROCESSED_REQUESTS_DELAY = timedelta(seconds=10)
    """Delay threshold to assume consistency of queue head operations after queue modifications."""

    _MAX_CACHED_REQUESTS = 1_000_000
    """Maximum number of requests that can be cached."""

    _MAX_HEAD_LIMIT = 1000
    """Cap on requests in progress when querying queue head."""

    _MAX_QUERIES_FOR_CONSISTENCY = 6
    """Maximum attempts to fetch a consistent queue head."""

    _QUERY_HEAD_BUFFER = 3
    """Multiplier for determining the number of requests to fetch based on in-progress requests."""

    _QUERY_HEAD_MIN_LENGTH = 100
    """The minimum number of requests fetched when querying the queue head."""

    _RECENTLY_HANDLED_CACHE_SIZE = 1000
    """Cache size for recently handled requests."""

    _STORAGE_CONSISTENCY_DELAY = timedelta(seconds=3)
    """Expected delay for storage to achieve consistency, guiding the timing of subsequent read operations."""

    def __init__(
        self,
        id: str,
        name: str | None,
        configuration: Configuration,
        client: BaseStorageClient,
    ) -> None:
        self._id = id
        self._name = name
        self._configuration = configuration

        # Get resource clients from storage client
        self._resource_client = client.request_queue(self._id)
        self._resource_collection_client = client.request_queues()

        # Other internal attributes
        self._tasks = list[asyncio.Task]()
        self._client_key = crypto_random_object_id()
        self._internal_timeout_seconds = 5 * 60
        self._assumed_total_count = 0
        self._assumed_handled_count = 0
        self._queue_head_dict: OrderedDictType[str, str] = OrderedDict()
        self._query_queue_head_task: asyncio.Task | None = None
        self._in_progress: set[str] = set()
        self._last_activity = datetime.now(timezone.utc)
        self._recently_handled: LRUCache[bool] = LRUCache(max_length=self._RECENTLY_HANDLED_CACHE_SIZE)
        self._requests_cache: LRUCache[dict] = LRUCache(max_length=self._MAX_CACHED_REQUESTS)

    @override
    @property
    def id(self) -> str:
        return self._id

    @override
    @property
    def name(self) -> str | None:
        return self._name

    @override
    @classmethod
    async def open(
        cls,
        *,
        id: str | None = None,
        name: str | None = None,
        configuration: Configuration | None = None,
    ) -> RequestQueue:
        from crawlee.storages._creation_management import open_storage

        return await open_storage(
            storage_class=cls,
            id=id,
            name=name,
            configuration=configuration,
        )

    @override
    async def drop(self, *, timeout: timedelta | None = None) -> None:
        from crawlee.storages._creation_management import remove_storage_from_cache

        # Wait for all tasks to finish
        await wait_for_all_tasks_for_finish(self._tasks, logger=logger, timeout=timeout)

        # Delete the storage from the underlying client and remove it from the cache
        await self._resource_client.delete()
        remove_storage_from_cache(storage_class=self.__class__, id=self._id, name=self._name)

    async def add_request(
        self,
        request: Request | BaseRequestData | str,
        *,
        forefront: bool = False,
    ) -> ProcessedRequest:
        """Adds a request to the `RequestQueue` while managing deduplication and positioning within the queue.

        The deduplication of requests relies on the `uniqueKey` field within the request dictionary. If `uniqueKey`
        exists, it remains unchanged; if it does not, it is generated based on the request's `url`, `method`,
        and `payload` fields. The generation of `uniqueKey` can be influenced by the `keep_url_fragment` and
        `use_extended_unique_key` flags, which dictate whether to include the URL fragment and the request's method
        and payload, respectively, in its computation.

        The request can be added to the forefront (beginning) or the back of the queue based on the `forefront`
        parameter. Information about the request's addition to the queue, including whether it was already present or
        handled, is returned in an output dictionary.

        Args:
            request: The request object to be added to the queue. Must include at least the `url` key.
                Optionaly it can include the `method`, `payload` and `uniqueKey` keys.

            forefront: If True, adds the request to the forefront of the queue; otherwise, adds it to the end.

            keep_url_fragment: Determines whether the URL fragment (the part of the URL after '#') should be retained
                in the unique key computation.

            use_extended_unique_key: Determines whether to use an extended unique key, incorporating the request's
                method and payload into the unique key computation.

        Returns: A dictionary containing information about the operation, including:
            - `requestId` The ID of the request.
            - `uniqueKey` The unique key associated with the request.
            - `wasAlreadyPresent` (bool): Indicates whether the request was already in the queue.
            - `wasAlreadyHandled` (bool): Indicates whether the request was already processed.
        """
        request = self._transform_request(request)
        self._last_activity = datetime.now(timezone.utc)

        cache_key = unique_key_to_request_id(request.unique_key)
        cached_info = self._requests_cache.get(cache_key)

        if cached_info:
            request.id = cached_info['id']
            # We may assume that if request is in local cache then also the information if the request was already
            # handled is there because just one client should be using one queue.
            return ProcessedRequest(
                id=request.id,
                unique_key=request.unique_key,
                was_already_present=True,
                was_already_handled=cached_info['was_already_handled'],
            )

        processed_request = await self._resource_client.add_request(request, forefront=forefront)
        processed_request.unique_key = request.unique_key

        self._cache_request(cache_key, processed_request)

        request_id, was_already_present = processed_request.id, processed_request.was_already_present
        is_handled = request.handled_at is not None

        if (
            not is_handled
            and not was_already_present
            and request_id not in self._in_progress
            and self._recently_handled.get(request_id) is None
        ):
            self._assumed_total_count += 1
            self._maybe_add_request_to_queue_head(request_id, forefront=forefront)

        return processed_request

    @override
    async def add_requests_batched(
        self,
        requests: Sequence[str | BaseRequestData | Request],
        *,
        batch_size: int = 1000,
        wait_time_between_batches: timedelta = timedelta(seconds=1),
        wait_for_all_requests_to_be_added: bool = False,
        wait_for_all_requests_to_be_added_timeout: timedelta | None = None,
    ) -> None:
        transformed_requests = self._transform_requests(requests)
        wait_time_secs = wait_time_between_batches.total_seconds()

        async def _process_batch(batch: Sequence[Request]) -> None:
            request_count = len(batch)
            response = await self._resource_client.batch_add_requests(batch)
            self._assumed_total_count += request_count
            logger.debug(f'Added {request_count} requests to the queue, response: {response}')

        # Wait for the first batch to be added
        first_batch = transformed_requests[:batch_size]
        if first_batch:
            await _process_batch(first_batch)

        async def _process_remaining_batches() -> None:
            for i in range(batch_size, len(transformed_requests), batch_size):
                batch = transformed_requests[i : i + batch_size]
                await _process_batch(batch)
                if i + batch_size < len(transformed_requests):
                    await asyncio.sleep(wait_time_secs)

        # Create and start the task to process remaining batches in the background
        remaining_batches_task = asyncio.create_task(_process_remaining_batches())
        self._tasks.append(remaining_batches_task)

        # Wait for all tasks to finish if requested
        if wait_for_all_requests_to_be_added:
            await wait_for_all_tasks_for_finish(
                self._tasks,
                logger=logger,
                timeout=wait_for_all_requests_to_be_added_timeout,
            )

    async def get_request(self, request_id: str) -> Request | None:
        """Retrieve a request from the queue.

        Args:
            request_id: ID of the request to retrieve.

        Returns:
            The retrieved request, or `None`, if it does not exist.
        """
        return await self._resource_client.get_request(request_id)

    async def fetch_next_request(self) -> Request | None:
        """Return the next request in the queue to be processed.

        Once you successfully finish processing of the request, you need to call `RequestQueue.mark_request_as_handled`
        to mark the request as handled in the queue. If there was some error in processing the request, call
        `RequestQueue.reclaim_request` instead, so that the queue will give the request to some other consumer
        in another call to the `fetch_next_request` method.

        Note that the `None` return value does not mean the queue processing finished, it means there are currently
        no pending requests. To check whether all requests in queue were finished, use `RequestQueue.is_finished`
        instead.

        Returns:
            The request or `None` if there are no more pending requests.
        """
        await self.ensure_head_is_non_empty()

        # We are likely done at this point.
        if len(self._queue_head_dict) == 0:
            return None

        next_request_id, _ = self._queue_head_dict.popitem(last=False)  # ~removeFirst()

        # This should never happen, but...
        if next_request_id in self._in_progress or self._recently_handled.get(next_request_id):
            logger.warning(
                'Queue head returned a request that is already in progress?!',
                extra={
                    'nextRequestId': next_request_id,
                    'inProgress': next_request_id in self._in_progress,
                    'recentlyHandled': next_request_id in self._recently_handled,
                },
            )
            return None
        self._in_progress.add(next_request_id)
        self._last_activity = datetime.now(timezone.utc)

        try:
            request = await self.get_request(next_request_id)
        except Exception:
            # On error, remove the request from in progress, otherwise it would be there forever
            self._in_progress.remove(next_request_id)
            raise

        # NOTE: It can happen that the queue head index is inconsistent with the main queue table.
        # This can occur in two situations:

        # 1)
        # Queue head index is ahead of the main table and the request is not present in the main table yet
        # (i.e. getRequest() returned null). In this case, keep the request marked as in progress for a short while,
        # so that isFinished() doesn't return true and _ensureHeadIsNonEmpty() doesn't not load the request into
        # the queueHeadDict straight again. After the interval expires, fetchNextRequest() will try to fetch this
        # request again, until it eventually appears in the main table.
        if request is None:
            logger.debug(
                'Cannot find a request from the beginning of queue, will be retried later',
                extra={'nextRequestId': next_request_id},
            )
            asyncio.get_running_loop().call_later(
                self._STORAGE_CONSISTENCY_DELAY.total_seconds(),
                lambda: self._in_progress.remove(next_request_id),
            )
            return None

        # 2)
        # Queue head index is behind the main table and the underlying request was already handled (by some other
        # client, since we keep the track of handled requests in recentlyHandled dictionary). We just add the request
        # to the recentlyHandled dictionary so that next call to _ensureHeadIsNonEmpty() will not put the request again
        # to queueHeadDict.
        if request.handled_at is not None:
            logger.debug(
                'Request fetched from the beginning of queue was already handled',
                extra={'nextRequestId': next_request_id},
            )
            self._recently_handled[next_request_id] = True
            return None

        return request

    async def mark_request_as_handled(self, request: Request) -> ProcessedRequest | None:
        """Mark a request as handled after successful processing.

        Handled requests will never again be returned by the `RequestQueue.fetch_next_request` method.

        Args:
            request: The request to mark as handled.

        Returns:
            Information about the queue operation. `None` if the given request was not in progress.
        """
        self._last_activity = datetime.now(timezone.utc)

        if request.id not in self._in_progress:
            logger.debug(f'Cannot mark request (ID: {request.id}) as handled, because it is not in progress!')
            return None

        if request.handled_at is None:
            request.handled_at = datetime.now(timezone.utc)

        processed_request = await self._resource_client.update_request(request)
        processed_request.unique_key = request.unique_key

        self._in_progress.remove(request.id)
        self._recently_handled[request.id] = True

        if not processed_request.was_already_handled:
            self._assumed_handled_count += 1

        self._cache_request(unique_key_to_request_id(request.unique_key), processed_request)
        return processed_request

    async def reclaim_request(
        self,
        request: Request,
        *,
        forefront: bool = False,
    ) -> ProcessedRequest | None:
        """Reclaim a failed request back to the queue.

        The request will be returned for processing later again by another call to `RequestQueue.fetchNextRequest`.

        Args:
            request: The request to return to the queue.
            forefront: Whether to add the request to the head or the end of the queue

        Returns:
            Information about the queue operation. `None` if the given request was not in progress.
        """
        self._last_activity = datetime.now(timezone.utc)

        if request.id not in self._in_progress:
            logger.debug(f'Cannot reclaim request (ID: {request.id}), because it is not in progress!')
            return None

        # TODO: If request hasn't been changed since the last getRequest(), we don't need to call updateRequest()
        # and thus improve performance.
        # https://github.com/apify/apify-sdk-python/issues/143
        processed_request = await self._resource_client.update_request(request, forefront=forefront)
        processed_request.unique_key = request.unique_key
        self._cache_request(unique_key_to_request_id(request.unique_key), processed_request)

        # Wait a little to increase a chance that the next call to fetchNextRequest() will return the request with
        # updated data. This is to compensate for the limitation of DynamoDB, where writes might not be immediately
        # visible to subsequent reads.
        def callback() -> None:
            if request.id not in self._in_progress:
                logger.debug(f'The request (ID: {request.id}) is no longer marked as in progress in the queue?!')
                return

            self._in_progress.remove(request.id)
            # Performance optimization: add request straight to head if possible
            self._maybe_add_request_to_queue_head(request.id, forefront=forefront)

        asyncio.get_running_loop().call_later(self._STORAGE_CONSISTENCY_DELAY.total_seconds(), callback)
        return processed_request

    async def is_empty(self) -> bool:
        """Check whether the queue is empty.

        Returns:
            bool: `True` if the next call to `RequestQueue.fetchNextRequest` would return `None`, otherwise `False`.
        """
        await self.ensure_head_is_non_empty()
        return len(self._queue_head_dict) == 0

    async def is_finished(self) -> bool:
        """Check whether the queue is finished.

        Due to the nature of distributed storage used by the queue,
        the function might occasionally return a false negative,
        but it will never return a false positive.

        Returns:
            bool: `True` if all requests were already handled and there are no more left. `False` otherwise.
        """
        seconds_since_last_activity = (datetime.now(timezone.utc) - self._last_activity).total_seconds()
        if self._in_progress_count() > 0 and seconds_since_last_activity > self._internal_timeout_seconds:
            message = (
                f'The request queue seems to be stuck for {self._internal_timeout_seconds}s, resetting internal state.'
            )
            logger.warning(message)
            self._reset()

        if len(self._queue_head_dict) > 0 or self._in_progress_count() > 0:
            return False

        # TODO: set ensure_consistency to True once the following issue is resolved:
        # https://github.com/apify/crawlee-python/issues/203
        is_head_consistent = await self.ensure_head_is_non_empty(ensure_consistency=False)
        return is_head_consistent and len(self._queue_head_dict) == 0 and self._in_progress_count() == 0

    async def get_info(self) -> RequestQueueMetadata | None:
        """Get an object containing general information about the request queue."""
        return await self._resource_client.get()

    @override
    async def get_handled_count(self) -> int:
        return self._assumed_handled_count

    @override
    async def get_total_count(self) -> int:
        return self._assumed_total_count

    async def ensure_head_is_non_empty(
        self,
        *,
        ensure_consistency: bool = False,
        limit: int | None = None,
        iteration: int = 0,
    ) -> bool:
        """Ensure that the queue head is non-empty.

        The method ensures that the queue head contains items. It may request more items than are currently
        in progress to guarantee that at least one item is present in the head of the queue.

        Args:
            ensure_consistency: If True, the query for the queue head is retried until the queue_modified_at is older
                than query_started_at by at least API_PROCESSED_REQUESTS_DELAY to ensure that the queue head is
                consistent.
            limit: The maximum number of items to fetch from the queue.
            iteration: To manage the recursion depth.

        Returns:
            True if the queue head is non-empty and consistent, False otherwise.
        """
        # If queue head is non-empty, returns True immediately
        if len(self._queue_head_dict) > 0:
            return True

        if limit is None:
            limit = max(self._in_progress_count() * self._QUERY_HEAD_BUFFER, self._QUERY_HEAD_MIN_LENGTH)

        if self._query_queue_head_task is None:
            self._query_queue_head_task = asyncio.Task(self._queue_query_head(limit))

        queue_head: RequestQueueHeadState = await self._query_queue_head_task

        # TODO: I feel this code below can be greatly simplified... (comes from TS implementation *wink*)
        # https://github.com/apify/apify-sdk-python/issues/142

        # If queue is still empty then one of the following holds:
        # - the other calls waiting for this task already consumed all the returned requests
        # - the limit was too low and contained only requests in progress
        # - the writes from other clients were not propagated yet
        # - the whole queue was processed and we are done

        # If limit was not reached in the call then there are no more requests to be returned.
        if queue_head.prev_limit >= self._MAX_HEAD_LIMIT:
            logger.warning(f'Reached the maximum number of requests in progress (limit: {self._MAX_HEAD_LIMIT})')

        should_repeat_with_higher_limit = (
            len(self._queue_head_dict) == 0
            and queue_head.was_limit_reached
            and queue_head.prev_limit < self._MAX_HEAD_LIMIT
        )

        # If ensure_consistency is True, we must ensure the database is consistent. It can be ensured if either:
        # - queue_modified_at is older than query_started_at by at least _API_PROCESSED_REQUESTS_DELAY
        # - had_multiple_clients is False and _assumed_total_count is less than _assumed_handled_count
        queue_latency = queue_head.query_started_at - queue_head.queue_modified_at.replace(tzinfo=timezone.utc)
        is_database_consistent = queue_latency.total_seconds() >= self._API_PROCESSED_REQUESTS_DELAY.total_seconds()

        is_locally_consistent = (
            not queue_head.had_multiple_clients and self._assumed_total_count <= self._assumed_handled_count
        )

        # Consistent information from one source is enough to consider request queue finished.
        should_repeat_for_consistency = ensure_consistency and not is_database_consistent and not is_locally_consistent

        # If both are false then head is consistent and we may exit.
        if not should_repeat_with_higher_limit and not should_repeat_for_consistency:
            return True

        # If we are querying for consistency then we limit the number of queries to MAX_QUERIES_FOR_CONSISTENCY.
        # If this is reached then we return false so that empty() and finished() returns possibly false negative.
        if not should_repeat_with_higher_limit and iteration > self._MAX_QUERIES_FOR_CONSISTENCY:
            return False

        next_limit = round(queue_head.prev_limit * 1.5) if should_repeat_with_higher_limit else queue_head.prev_limit

        # If we are repeating for consistency then wait required time.
        if should_repeat_for_consistency:
            elapsed_time = (datetime.now(timezone.utc) - queue_head.queue_modified_at).total_seconds()
            delay_seconds = self._API_PROCESSED_REQUESTS_DELAY.total_seconds() - elapsed_time
            logger.info(f'Waiting for {delay_seconds} for queue finalization, to ensure data consistency.')
            await asyncio.sleep(delay_seconds)

        return await self.ensure_head_is_non_empty(
            ensure_consistency=ensure_consistency,
            limit=next_limit,
            iteration=iteration + 1,
        )

    def _in_progress_count(self) -> int:
        return len(self._in_progress)

    def _reset(self) -> None:
        self._queue_head_dict.clear()
        self._query_queue_head_task = None
        self._in_progress.clear()
        self._recently_handled.clear()
        self._assumed_total_count = 0
        self._assumed_handled_count = 0
        self._requests_cache.clear()
        self._last_activity = datetime.now(timezone.utc)

    def _cache_request(self, cache_key: str, processed_request: ProcessedRequest) -> None:
        self._requests_cache[cache_key] = {
            'id': processed_request.id,
            'is_handled': processed_request.was_already_handled,
            'unique_key': processed_request.unique_key,
            'was_already_handled': processed_request.was_already_handled,
        }

    async def _queue_query_head(self, limit: int) -> RequestQueueHeadState:
        query_started_at = datetime.now(timezone.utc)

        list_head = await self._resource_client.list_head(limit=limit)
        list_head_items: list[Request] = list_head.items

        for request in list_head_items:
            # Queue head index might be behind the main table, so ensure we don't recycle requests
            if (
                not request.id
                or not request.unique_key
                or request.id in self._in_progress
                or self._recently_handled.get(request.id)
            ):
                continue

            self._queue_head_dict[request.id] = request.id
            self._cache_request(
                cache_key=unique_key_to_request_id(request.unique_key),
                processed_request=ProcessedRequest(
                    id=request.id,
                    unique_key=request.unique_key,
                    was_already_handled=False,
                    was_already_present=True,
                ),
            )

        # This is needed so that the next call to _ensureHeadIsNonEmpty() will fetch the queue head again.
        self._query_queue_head_task = None

        return RequestQueueHeadState(
            was_limit_reached=len(list_head.items) >= limit,
            prev_limit=limit,
            queue_modified_at=list_head.queue_modified_at,
            query_started_at=query_started_at,
            had_multiple_clients=list_head.had_multiple_clients,
        )

    def _maybe_add_request_to_queue_head(
        self,
        request_id: str,
        *,
        forefront: bool,
    ) -> None:
        if forefront:
            self._queue_head_dict[request_id] = request_id
            # Move to start, i.e. forefront of the queue
            self._queue_head_dict.move_to_end(request_id, last=False)
        elif self._assumed_total_count < self._QUERY_HEAD_MIN_LENGTH:
            # OrderedDict puts the item to the end of the queue by default
            self._queue_head_dict[request_id] = request_id



# ruff: noqa: TRY301, FBT002, UP007
from __future__ import annotations

from pathlib import Path
from typing import Annotated, Optional, cast

import httpx
import inquirer  # type: ignore
import typer
from cookiecutter.main import cookiecutter  # type: ignore
from inquirer.render.console import ConsoleRender  # type: ignore
from rich.progress import Progress, SpinnerColumn, TextColumn

TEMPLATE_LIST_URL = 'https://api.github.com/repos/apify/crawlee-python/contents/templates'

cli = typer.Typer(no_args_is_help=True)


@cli.callback(invoke_without_command=True)
def callback(
    version: Annotated[
        bool,
        typer.Option(
            '-V',
            '--version',
            is_flag=True,
            help='Print Crawlee version',
        ),
    ] = False,
) -> None:
    """Crawlee is a web scraping and browser automation library."""
    if version:
        from crawlee import __version__

        typer.echo(__version__)


def _prompt_for_project_name(initial_project_name: str | None) -> str:
    """Prompt the user for a non-empty project name that does not lead to an existing folder."""
    while True:
        if initial_project_name is not None:
            project_name = initial_project_name
            initial_project_name = None
        else:
            project_name = ConsoleRender().render(
                inquirer.Text(
                    name='project_name',
                    message='Name of the new project folder',
                    validate=lambda _, value: bool(value.strip()),
                ),
            )

        if not project_name:
            typer.echo('Project name is required.', err=True)
            continue

        project_path = Path.cwd() / project_name

        if project_path.exists():
            typer.echo(f'Folder {project_path} already exists. Please choose another name.', err=True)
            continue

        return project_name


def _prompt_for_template() -> str:
    """Prompt the user to select a template from a list."""
    # Fetch available templates
    response = httpx.get(TEMPLATE_LIST_URL, timeout=httpx.Timeout(10))
    response.raise_for_status()
    template_choices = [item['name'] for item in response.json() if item['type'] == 'dir']

    # Prompt for template choice
    return cast(
        str,
        ConsoleRender().render(
            inquirer.List(
                name='template',
                message='Please select the template for your new Crawlee project',
                choices=[(choice[0].upper() + choice[1:], choice) for choice in template_choices],
            ),
        ),
    )


@cli.command()
def create(
    project_name: Optional[str] = typer.Argument(
        default=None,
        help='The name of the project and the directory that will be created to contain it. '
        'If none is given, you will be prompted.',
        show_default=False,
    ),
    template: Optional[str] = typer.Option(
        default=None,
        help='The template to be used to create the project. If none is given, you will be prompted.',
        show_default=False,
    ),
) -> None:
    """Bootstrap a new Crawlee project."""
    try:
        # Prompt for project name if not provided.
        project_name = _prompt_for_project_name(project_name)

        # Prompt for template choice if not provided.
        if template is None:
            template = _prompt_for_template()

        if project_name and template:
            # Start the bootstrap process.
            with Progress(
                SpinnerColumn(),
                TextColumn('[progress.description]{task.description}'),
                transient=True,
            ) as progress:
                progress.add_task(description='Bootstrapping...', total=None)
                cookiecutter(
                    template='gh:apify/crawlee-python',
                    directory=f'templates/{template}',
                    no_input=True,
                    extra_context={'project_name': project_name},
                )

            typer.echo(f'Your project "{project_name}" was created.')
            typer.echo(
                f'To run it, navigate to the directory: "cd {project_name}", '
                'install dependencies with "poetry install", '
                f'and run it using "poetry run python -m {project_name}".'
            )
            typer.echo(f'See the "{project_name}/README.md" for more information.')

    except httpx.HTTPStatusError as exc:
        typer.echo(f'Failed to fetch templates: {exc}.', err=True)
    except KeyboardInterrupt:
        typer.echo('Operation cancelled by user.')



from __future__ import annotations

from typing import Any, AsyncGenerator, Awaitable, Callable, Generator, Generic, cast

from typing_extensions import TypeVar

from crawlee.basic_crawler.errors import (
    ContextPipelineFinalizationError,
    ContextPipelineInitializationError,
    ContextPipelineInterruptedError,
    RequestHandlerError,
    SessionError,
)
from crawlee.basic_crawler.types import BasicCrawlingContext

TCrawlingContext = TypeVar('TCrawlingContext', bound=BasicCrawlingContext, default=BasicCrawlingContext)
TMiddlewareCrawlingContext = TypeVar('TMiddlewareCrawlingContext', bound=BasicCrawlingContext)


class ContextPipeline(Generic[TCrawlingContext]):
    """Encapsulates the logic of gradually enhancing the crawling context with additional information and utilities.

    The enhancement is done by a chain of middlewares that are added to the pipeline after it's creation.
    """

    def __init__(
        self,
        *,
        _middleware: Callable[
            [TCrawlingContext],
            AsyncGenerator[TMiddlewareCrawlingContext, None],
        ]
        | None = None,
        _parent: ContextPipeline[BasicCrawlingContext] | None = None,
    ) -> None:
        self._middleware = _middleware
        self._parent = _parent

    def _middleware_chain(self) -> Generator[ContextPipeline[Any], None, None]:
        yield self

        if self._parent is not None:
            yield from self._parent._middleware_chain()  # noqa: SLF001

    async def __call__(
        self,
        crawling_context: BasicCrawlingContext,
        final_context_consumer: Callable[[TCrawlingContext], Awaitable[None]],
    ) -> None:
        """Run a crawling context through the middleware chain and pipe it into a consumer function.

        Exceptions from the consumer function are wrapped together with the final crawling context.
        """
        chain = list(self._middleware_chain())
        cleanup_stack = list[AsyncGenerator]()

        try:
            for member in reversed(chain):
                if member._middleware:  # noqa: SLF001
                    middleware_instance = member._middleware(crawling_context)  # noqa: SLF001
                    try:
                        result = await middleware_instance.__anext__()
                    except SessionError:  # Session errors get special treatment
                        raise
                    except StopAsyncIteration as e:
                        raise RuntimeError('The middleware did not yield') from e
                    except ContextPipelineInterruptedError:
                        raise
                    except Exception as e:
                        raise ContextPipelineInitializationError(e, crawling_context) from e

                    crawling_context = result
                    cleanup_stack.append(middleware_instance)

            try:
                await final_context_consumer(cast(TCrawlingContext, crawling_context))
            except SessionError:  # Session errors get special treatment
                raise
            except Exception as e:
                raise RequestHandlerError(e, crawling_context) from e
        finally:
            for middleware_instance in reversed(cleanup_stack):
                try:
                    result = await middleware_instance.__anext__()
                except StopAsyncIteration:  # noqa: PERF203
                    pass
                except ContextPipelineInterruptedError as e:
                    raise RuntimeError('Invalid state - pipeline interrupted in the finalization step') from e
                except Exception as e:
                    raise ContextPipelineFinalizationError(e, crawling_context) from e
                else:
                    raise RuntimeError('The middleware yielded more than once')

    def compose(
        self,
        middleware: Callable[
            [TCrawlingContext],
            AsyncGenerator[TMiddlewareCrawlingContext, None],
        ],
    ) -> ContextPipeline[TMiddlewareCrawlingContext]:
        """Add a middleware to the pipeline.

        The middleware should yield exactly once, and it should yield an (optionally) extended crawling context object.
        The part before the yield can be used for initialization and the part after it for cleanup.

        Returns:
            The extended pipeline instance, providing a fluent interface
        """
        return ContextPipeline[TMiddlewareCrawlingContext](
            _middleware=cast(
                Callable[[BasicCrawlingContext], AsyncGenerator[TMiddlewareCrawlingContext, None]], middleware
            ),
            _parent=cast(ContextPipeline[BasicCrawlingContext], self),
        )



# Inspiration: https://github.com/apify/crawlee/blob/v3.8.2/packages/types/src/storages.ts#L314:L328

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .base_dataset_client import BaseDatasetClient
    from .base_dataset_collection_client import BaseDatasetCollectionClient
    from .base_key_value_store_client import BaseKeyValueStoreClient
    from .base_key_value_store_collection_client import BaseKeyValueStoreCollectionClient
    from .base_request_queue_client import BaseRequestQueueClient
    from .base_request_queue_collection_client import BaseRequestQueueCollectionClient


class BaseStorageClient(ABC):
    """Defines an abstract base for storage clients.

    It offers interfaces to get subclients for interacting with storage resources like datasets, key-value stores,
    and request queues.
    """

    @abstractmethod
    def dataset(self, id: str) -> BaseDatasetClient:
        """Gets a subclient for a specific dataset by its ID."""

    @abstractmethod
    def datasets(self) -> BaseDatasetCollectionClient:
        """Gets a subclient for dataset collection operations."""

    @abstractmethod
    def key_value_store(self, id: str) -> BaseKeyValueStoreClient:
        """Gets a subclient for a specific key-value store by its ID."""

    @abstractmethod
    def key_value_stores(self) -> BaseKeyValueStoreCollectionClient:
        """Gets a subclient for key-value store collection operations."""

    @abstractmethod
    def request_queue(self, id: str) -> BaseRequestQueueClient:
        """Gets a subclient for a specific request queue by its ID."""

    @abstractmethod
    def request_queues(self) -> BaseRequestQueueCollectionClient:
        """Gets a subclient for request queue collection operations."""

    @abstractmethod
    async def purge_on_start(self) -> None:
        """Performs a purge of the default storages.

        This method ensures that the purge is executed only once during the lifetime of the instance.
        It is primarily used to clean up residual data from previous runs to maintain a clean state.
        If the storage client does not support purging, leave it empty.
        """



from __future__ import annotations


def compute_weighted_avg(values: list[float], weights: list[float]) -> float:
    """Computes a weighted average of an array of numbers, complemented by an array of weights.

    Args:
        values: List of values.
        weights: List of weights.

    Raises:
        ValueError: If total weight is zero.

    Returns:
        float: Weighted average.
    """
    if not values or not weights:
        raise ValueError('Values and weights lists must not be empty')

    if len(values) != len(weights):
        raise ValueError('Values and weights must be of equal length')

    # zip(..., strict=True) can be used once support for Python 3.9 is dropped
    result = sum(value * weight for value, weight in zip(values, weights))
    total_weight = sum(weights)

    if total_weight == 0:
        raise ValueError('Total weight cannot be zero')

    return result / total_weight



from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime, timedelta, timezone
from typing import TYPE_CHECKING, Union

if TYPE_CHECKING:
    from crawlee._utils.byte_size import ByteSize


@dataclass
class LoadRatioInfo:
    """Represents the load ratio of a resource.

    Args:
        limit_ratio: The maximum ratio of overloaded and non-overloaded samples. If the actual ratio exceeds this
            value, the resource is considered as overloaded.

        actual_ratio: The actual ratio of overloaded and non-overloaded samples.
    """

    limit_ratio: float
    actual_ratio: float

    @property
    def is_overloaded(self) -> bool:
        """Returns whether the resource is overloaded."""
        return self.actual_ratio > self.limit_ratio


@dataclass
class SystemInfo:
    """Represents the current status of the system.

    Args:
        cpu_info: The CPU load ratio.
        memory_info: The memory load ratio.
        event_loop_info: The event loop load ratio.
        client_info: The client load ratio.
        created_at: The time at which the measurement was taken.
    """

    cpu_info: LoadRatioInfo
    memory_info: LoadRatioInfo
    event_loop_info: LoadRatioInfo
    client_info: LoadRatioInfo
    created_at: datetime = field(default_factory=lambda: datetime.now(timezone.utc))

    @property
    def is_system_idle(self) -> bool:
        """Indicates whether the system is currently idle or overloaded."""
        return (
            not self.cpu_info.is_overloaded
            and not self.memory_info.is_overloaded
            and not self.event_loop_info.is_overloaded
            and not self.client_info.is_overloaded
        )

    def __str__(self) -> str:
        """Get a string representation of the system info."""
        stats = {
            'cpu': self.cpu_info.actual_ratio,
            'mem': self.memory_info.actual_ratio,
            'event_loop': self.event_loop_info.actual_ratio,
            'client_info': self.client_info.actual_ratio,
        }
        return '; '.join(f'{name} = {ratio}' for name, ratio in stats.items())


@dataclass
class CpuSnapshot:
    """A snapshot of CPU usage.

    Args:
        used_ratio: The ratio of CPU currently in use.
        max_used_ratio: The maximum ratio of CPU that is considered acceptable.
        created_at: The time at which the measurement was taken.
    """

    used_ratio: float
    max_used_ratio: float
    created_at: datetime = field(default_factory=lambda: datetime.now(timezone.utc))

    @property
    def is_overloaded(self) -> bool:
        """Returns whether the CPU is considered as overloaded."""
        return self.used_ratio > self.max_used_ratio


@dataclass
class MemorySnapshot:
    """A snapshot of memory usage.

    Args:
        total_size: Total memory available in the system.
        current_size: Memory usage of the current Python process and its children.
        max_memory_size: The maximum memory that can be used by `AutoscaledPool`.
        max_used_memory_ratio: The maximum acceptable ratio of `current_size` to `max_memory_size`.
        created_at: The time at which the measurement was taken.
    """

    total_size: ByteSize
    current_size: ByteSize
    max_memory_size: ByteSize
    max_used_memory_ratio: float
    created_at: datetime = field(default_factory=lambda: datetime.now(timezone.utc))

    @property
    def is_overloaded(self) -> bool:
        """Returns whether the memory is considered as overloaded."""
        return (self.current_size / self.max_memory_size) > self.max_used_memory_ratio


@dataclass
class EventLoopSnapshot:
    """Snapshot of the state of the event loop.

    Args:
        delay: The current delay of the event loop.
        max_delay: The maximum delay that is considered acceptable.
        created_at: The time at which the measurement was taken.
    """

    delay: timedelta
    max_delay: timedelta
    created_at: datetime = field(default_factory=lambda: datetime.now(timezone.utc))

    @property
    def max_delay_exceeded(self) -> timedelta:
        """Returns the amount of time by which the delay exceeds the maximum delay."""
        return max(self.delay - self.max_delay, timedelta(seconds=0))

    @property
    def is_overloaded(self) -> bool:
        """Returns whether the event loop is considered as overloaded."""
        return self.delay > self.max_delay


@dataclass
class ClientSnapshot:
    """Snapshot of the state of the client.

    Args:
        error_count: The number of errors (HTTP 429) that occurred.
        max_error_count: The maximum number of errors that is considered acceptable.
        created_at: The time at which the measurement was taken.
    """

    error_count: int
    max_error_count: int
    created_at: datetime = field(default_factory=lambda: datetime.now(timezone.utc))

    @property
    def is_overloaded(self) -> bool:
        """Returns whether the client is considered as overloaded."""
        return self.error_count > self.max_error_count


Snapshot = Union[MemorySnapshot, CpuSnapshot, EventLoopSnapshot, ClientSnapshot]



from __future__ import annotations

import json
from enum import Enum
from typing import TYPE_CHECKING, Any, NoReturn

from crawlee._utils.file import ContentType, is_content_type

if TYPE_CHECKING:
    from crawlee.types import StorageTypes


def filter_out_none_values_recursively(dictionary: dict, *, remove_empty_dicts: bool = False) -> dict | None:
    """Recursively filters out None values from a dictionary.

    Args:
        dictionary: The dictionary to filter.
        remove_empty_dicts: Flag indicating whether to remove empty nested dictionaries.

    Returns:
        A copy of the dictionary with all None values (and potentially empty dictionaries) removed.
    """
    result = {}
    for k, v in dictionary.items():
        # If the value is a dictionary, apply recursion
        if isinstance(v, dict):
            nested = filter_out_none_values_recursively(v, remove_empty_dicts=remove_empty_dicts)
            if nested or not remove_empty_dicts:
                result[k] = nested
        elif v is not None:
            result[k] = v

    # If removing empty dictionaries and result is empty, return None
    if remove_empty_dicts and not result:
        return None
    return result


def maybe_extract_enum_member_value(maybe_enum_member: Any) -> Any:
    """Extract the value of an enumeration member if it is an Enum, otherwise return the original value."""
    if isinstance(maybe_enum_member, Enum):
        return maybe_enum_member.value
    return maybe_enum_member


def maybe_parse_body(body: bytes, content_type: str) -> Any:
    """Parse the response body based on the content type."""
    if is_content_type(ContentType.JSON, content_type):
        s = body.decode('utf-8')
        return json.loads(s)

    if is_content_type(ContentType.XML, content_type) or is_content_type(ContentType.TEXT, content_type):
        return body.decode('utf-8')

    return body


def raise_on_duplicate_storage(client_type: StorageTypes, key_name: str, value: str) -> NoReturn:
    """Raise an error indicating that a storage with the provided key name and value already exists."""
    client_type = maybe_extract_enum_member_value(client_type)
    raise ValueError(f'{client_type} with {key_name} "{value}" already exists.')


def raise_on_non_existing_storage(client_type: StorageTypes, id: str | None) -> NoReturn:
    """Raise an error indicating that a storage with the provided id does not exist."""
    client_type = maybe_extract_enum_member_value(client_type)
    raise ValueError(f'{client_type} with id "{id}" does not exist.')



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from crawlee.basic_crawler.types import BasicCrawlingContext, EnqueueLinksFunction

if TYPE_CHECKING:
    from playwright.async_api import Page, Response


@dataclass(frozen=True)
class PlaywrightCrawlingContext(BasicCrawlingContext):
    """Crawling context used by PlaywrightSoupCrawler.

    Args:
        page: The Playwright `Page` object.
        response: The Playwright `Response` object.
        enqueue_links: The `PlaywrightCrawler` implementation of the `EnqueueLinksFunction` function.
    """

    page: Page
    response: Response
    enqueue_links: EnqueueLinksFunction



# Inspiration: https://github.com/apify/crawlee/blob/v3.7.3/packages/core/src/events/event_manager.ts

from __future__ import annotations

import asyncio
from collections import defaultdict
from datetime import timedelta
from functools import wraps
from logging import getLogger
from typing import TYPE_CHECKING, TypedDict

from pyee.asyncio import AsyncIOEventEmitter
from typing_extensions import NotRequired

from crawlee._utils.recurring_task import RecurringTask
from crawlee._utils.wait import wait_for_all_tasks_for_finish
from crawlee.events.types import Event, EventPersistStateData

if TYPE_CHECKING:
    from types import TracebackType

    from crawlee.events.types import EventData, Listener, WrappedListener

logger = getLogger(__name__)


class EventManagerOptions(TypedDict):
    """Parameter types for subclass __init__ methods, copied from EventManager.__init__."""

    persist_state_interval: NotRequired[timedelta]
    close_timeout: NotRequired[timedelta | None]


class EventManager:
    """Event manager for registering, emitting, and managing event listeners.

    Event manager allows you to register event listeners, emit events, and wait for event listeners to complete
    their execution. It is built on top of the `pyee.asyncio.AsyncIOEventEmitter` class.
    """

    def __init__(
        self,
        *,
        persist_state_interval: timedelta = timedelta(minutes=1),
        close_timeout: timedelta | None = None,
    ) -> None:
        """Create a new instance.

        Args:
            persist_state_interval: Interval at which `PersistState` events are emitted.
            close_timeout: Optional timeout after which the pending event listeners are canceled.
        """
        self._persist_state_interval = persist_state_interval
        self._close_timeout = close_timeout

        # Asynchronous event emitter for handle events and invoke the event listeners.
        self._event_emitter = AsyncIOEventEmitter()

        # Listeners are wrapped inside asyncio.Task. Store their references here so that we can wait for them to finish.
        self._listener_tasks: set[asyncio.Task] = set()

        # Store the mapping between events, listeners and their wrappers in the following way:
        #   event -> listener -> [wrapped_listener_1, wrapped_listener_2, ...]
        self._listeners_to_wrappers: dict[Event, dict[Listener, list[WrappedListener]]] = defaultdict(
            lambda: defaultdict(list),
        )

        # Recurring task for emitting persist state events.
        self._emit_persist_state_event_rec_task = RecurringTask(
            func=self._emit_persist_state_event,
            delay=self._persist_state_interval,
        )

    async def __aenter__(self) -> EventManager:
        """Initializes the event manager upon entering the async context."""
        self._emit_persist_state_event_rec_task.start()
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        exc_traceback: TracebackType | None,
    ) -> None:
        """Closes the local event manager upon exiting the async context.

        This will stop listening for the events, and it will wait for all the event listeners to finish.
        """
        await self.wait_for_all_listeners_to_complete(timeout=self._close_timeout)
        self._event_emitter.remove_all_listeners()
        self._listener_tasks.clear()
        self._listeners_to_wrappers.clear()
        await self._emit_persist_state_event_rec_task.stop()

    def on(self, *, event: Event, listener: Listener) -> None:
        """Add an event listener to the event manager.

        Args:
            event: The Actor event for which to listen to.
            listener: The function (sync or async) which is to be called when the event is emitted.
        """

        @wraps(listener)
        async def listener_wrapper(event_data: EventData) -> None:
            # If the listener is a coroutine function, just call it, otherwise, run it in a separate thread
            # to avoid blocking the event loop
            coro = (
                listener(event_data)
                if asyncio.iscoroutinefunction(listener)
                else asyncio.to_thread(listener, event_data)
            )
            # Note: use `asyncio.iscoroutinefunction` rather then `inspect.iscoroutinefunction` since it works with
            # unittests.mock.AsyncMock. See https://github.com/python/cpython/issues/84753.

            listener_task = asyncio.create_task(coro, name=f'Task-{event.value}-{listener.__name__}')
            self._listener_tasks.add(listener_task)

            try:
                logger.debug('LocalEventManager.on.listener_wrapper(): Awaiting listener task...')
                await listener_task
                logger.debug('LocalEventManager.on.listener_wrapper(): Listener task completed.')
            except Exception:
                # We need to swallow the exception and just log it here, otherwise it could break the event emitter
                logger.exception(
                    'Exception in the event listener',
                    extra={'event_name': event.value, 'listener_name': listener.__name__},
                )
            finally:
                logger.debug('LocalEventManager.on.listener_wrapper(): Removing listener task from the set...')
                self._listener_tasks.remove(listener_task)

        self._listeners_to_wrappers[event][listener].append(listener_wrapper)
        self._event_emitter.add_listener(event.value, listener_wrapper)

    def off(self, *, event: Event, listener: Listener | None = None) -> None:
        """Remove a listener, or all listeners, from an Actor event.

        Args:
            event: The Actor event for which to remove listeners.
            listener: The listener which is supposed to be removed. If not passed, all listeners of this event
                are removed.
        """
        if listener:
            for listener_wrapper in self._listeners_to_wrappers[event][listener]:
                self._event_emitter.remove_listener(event.value, listener_wrapper)
            self._listeners_to_wrappers[event][listener] = []
        else:
            self._listeners_to_wrappers[event] = defaultdict(list)
            self._event_emitter.remove_all_listeners(event.value)

    def emit(self, *, event: Event, event_data: EventData) -> None:
        """Emit an event.

        Args:
            event: The event which will be emitted.
            event_data: The data which will be passed to the event listeners.
        """
        self._event_emitter.emit(event.value, event_data)

    async def wait_for_all_listeners_to_complete(self, *, timeout: timedelta | None = None) -> None:
        """Wait for all currently executing event listeners to complete.

        Args:
            timeout: The maximum time to wait for the event listeners to finish. If they do not complete within
                the specified timeout, they will be canceled.
        """

        async def wait_for_listeners() -> None:
            """Gathers all listener tasks and awaits their completion, logging any exceptions encountered."""
            results = await asyncio.gather(*self._listener_tasks, return_exceptions=True)
            for result in results:
                if isinstance(result, Exception):
                    logger.exception('Event listener raised an exception.', exc_info=result)

        tasks = [asyncio.create_task(wait_for_listeners(), name=f'Task-{wait_for_listeners.__name__}')]

        await wait_for_all_tasks_for_finish(tasks=tasks, logger=logger, timeout=timeout)

    async def _emit_persist_state_event(self) -> None:
        """Emits a persist state event with the given migration status."""
        self.emit(event=Event.PERSIST_STATE, event_data=EventPersistStateData(is_migrating=False))



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Literal

from typing_extensions import Unpack

from crawlee._utils.blocked import RETRY_CSS_SELECTORS
from crawlee.basic_crawler import BasicCrawler, BasicCrawlerOptions, ContextPipeline
from crawlee.basic_crawler.errors import SessionError
from crawlee.browsers import BrowserPool
from crawlee.enqueue_strategy import EnqueueStrategy
from crawlee.models import BaseRequestData
from crawlee.playwright_crawler.types import PlaywrightCrawlingContext

if TYPE_CHECKING:
    from collections.abc import AsyncGenerator

    from crawlee.basic_crawler.types import AddRequestsKwargs, BasicCrawlingContext


class PlaywrightCrawler(BasicCrawler[PlaywrightCrawlingContext]):
    """A crawler that leverages the [Playwright](https://playwright.dev/python/) browser automation library.

    `PlaywrightCrawler` is a subclass of `BasicCrawler`, inheriting all its features, such as autoscaling of requests,
    request routing, and utilization of `RequestProvider`. Additionally, it offers Playwright-specific methods and
    properties, like the `page` property for user data extraction, and the `enqueue_links` method for crawling
    other pages.

    This crawler is ideal for crawling websites that require JavaScript execution, as it uses headless browsers
    to download web pages and extract data. For websites that do not require JavaScript, consider using
    `BeautifulSoupCrawler`, which uses raw HTTP requests, and it is much faster.

    `PlaywrightCrawler` opens a new browser page (i.e., tab) for each `Request` object and invokes the user-provided
    request handler function via the `Router`. Users can interact with the page and extract the data using
    the Playwright API.

    Note that the pool of browser instances used by `PlaywrightCrawler`, and the pages they open, is internally
    managed by the `BrowserPool`.
    """

    def __init__(
        self,
        browser_pool: BrowserPool | None = None,
        browser_type: Literal['chromium', 'firefox', 'webkit'] | None = None,
        headless: bool | None = None,
        **kwargs: Unpack[BasicCrawlerOptions[PlaywrightCrawlingContext]],
    ) -> None:
        """Create a new instance.

        Args:
            browser_pool: A `BrowserPool` instance to be used for launching the browsers and getting pages.
            browser_type: The type of browser to launch ('chromium', 'firefox', or 'webkit').
                This option should not be used if `browser_pool` is provided.
            headless: Whether to run the browser in headless mode.
                This option should not be used if `browser_pool` is provided.
            kwargs: Additional arguments to be forwarded to the underlying `BasicCrawler`.
        """
        if browser_pool:
            # Raise an exception if browser_pool is provided together with headless or browser_type arguments.
            if headless is not None or browser_type is not None:
                raise ValueError(
                    'You cannot provide `headless` or `browser_type` arguments when `browser_pool` is provided.'
                )

        # If browser_pool is not provided, create a new instance of BrowserPool with specified arguments.
        else:
            browser_pool = BrowserPool.with_default_plugin(headless=headless, browser_type=browser_type)

        self._browser_pool = browser_pool

        # Compose the context pipeline with the Playwright-specific context enhancer.
        kwargs['_context_pipeline'] = (
            ContextPipeline().compose(self._make_http_request).compose(self._handle_blocked_request)
        )
        kwargs['_additional_context_managers'] = [self._browser_pool]
        kwargs.setdefault('_logger', logging.getLogger(__name__))

        super().__init__(**kwargs)

    async def _make_http_request(
        self,
        context: BasicCrawlingContext,
    ) -> AsyncGenerator[PlaywrightCrawlingContext, None]:
        """Enhance the crawling context with making an HTTP request using Playwright.

        Args:
            context: The basic crawling context to be enhanced.

        Raises:
            ValueError: If the browser pool is not initialized.
            SessionError: If the URL cannot be loaded by the browser.

        Yields:
            An enhanced crawling context with Playwright-specific features.
        """
        if self._browser_pool is None:
            raise ValueError('Browser pool is not initialized.')

        # Create a new browser page, navigate to the URL and get response.
        crawlee_page = await self._browser_pool.new_page(proxy_info=context.proxy_info)
        response = await crawlee_page.page.goto(context.request.url)

        if response is None:
            raise SessionError(f'Failed to load the URL: {context.request.url}')

        # Set the loaded URL to the actual URL after redirection.
        context.request.loaded_url = crawlee_page.page.url

        async def enqueue_links(
            *,
            selector: str = 'a',
            label: str | None = None,
            user_data: dict | None = None,
            **kwargs: Unpack[AddRequestsKwargs],
        ) -> None:
            """The `PlaywrightCrawler` implementation of the `EnqueueLinksFunction` function."""
            kwargs.setdefault('strategy', EnqueueStrategy.SAME_HOSTNAME)

            requests = list[BaseRequestData]()
            user_data = user_data or {}

            elements = await crawlee_page.page.query_selector_all(selector)

            for element in elements:
                href = await element.get_attribute('href')

                if href:
                    link_user_data = user_data.copy()

                    if label is not None:
                        link_user_data.setdefault('label', label)

                    request = BaseRequestData.from_url(href, user_data=link_user_data)
                    requests.append(request)

            await context.add_requests(requests, **kwargs)

        yield PlaywrightCrawlingContext(
            request=context.request,
            session=context.session,
            add_requests=context.add_requests,
            send_request=context.send_request,
            push_data=context.push_data,
            proxy_info=context.proxy_info,
            log=context.log,
            page=crawlee_page.page,
            response=response,
            enqueue_links=enqueue_links,
        )

        await crawlee_page.page.close()

    async def _handle_blocked_request(
        self,
        crawling_context: PlaywrightCrawlingContext,
    ) -> AsyncGenerator[PlaywrightCrawlingContext, None]:
        """Enhance the crawling context with handling of blocked requests.

        Args:
            crawling_context: The crawling context to be checked for blocking.

        Raises:
            SessionError: If the session is blocked based on the HTTP status code or the response content.

        Yields:
            The original crawling context if the session is not blocked.
        """
        if self._retry_on_blocked:
            status_code = crawling_context.response.status

            # Check if the session is blocked based on the HTTP status code.
            if crawling_context.session and crawling_context.session.is_blocked_status_code(status_code=status_code):
                raise SessionError(f'Assuming the session is blocked based on HTTP status code {status_code}.')

            matched_selectors = [
                selector for selector in RETRY_CSS_SELECTORS if (await crawling_context.page.query_selector(selector))
            ]

            # Check if the session is blocked based on the response content
            if matched_selectors:
                raise SessionError(
                    'Assuming the session is blocked - '
                    f"HTTP response matched the following selectors: {'; '.join(matched_selectors)}"
                )

        yield crawling_context



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, AsyncContextManager

if TYPE_CHECKING:
    from httpx import Response

    from crawlee.models import KeyValueStoreListKeysPage, KeyValueStoreMetadata, KeyValueStoreRecord


class BaseKeyValueStoreClient(ABC):
    """Abstract base class for key-value store resource clients.

    These clients are specific to the type of resource they manage and operate under a designated storage
    client, like a memory storage client.
    """

    @abstractmethod
    async def get(self) -> KeyValueStoreMetadata | None:
        """Get metadata about the key-value store being managed by this client.

        Returns:
            An object containing the key-value store's details, or None if the key-value store does not exist.
        """

    @abstractmethod
    async def update(
        self,
        *,
        name: str | None = None,
    ) -> KeyValueStoreMetadata:
        """Update the key-value store metadata.

        Args:
            name: New new name for the key-value store.

        Returns:
            An object reflecting the updated key-value store metadata.
        """

    @abstractmethod
    async def delete(self) -> None:
        """Permanently delete the key-value store managed by this client."""

    @abstractmethod
    async def list_keys(
        self,
        *,
        limit: int = 1000,
        exclusive_start_key: str | None = None,
    ) -> KeyValueStoreListKeysPage:
        """List the keys in the key-value store.

        Args:
            limit: Number of keys to be returned. Maximum value is 1000.
            exclusive_start_key: All keys up to this one (including) are skipped from the result.

        Returns:
            The list of keys in the key-value store matching the given arguments.
        """

    @abstractmethod
    async def get_record(self, key: str) -> KeyValueStoreRecord | None:
        """Retrieve the given record from the key-value store.

        Args:
            key: Key of the record to retrieve

        Returns:
            The requested record, or None, if the record does not exist
        """

    @abstractmethod
    async def get_record_as_bytes(self, key: str) -> KeyValueStoreRecord[bytes] | None:
        """Retrieve the given record from the key-value store, without parsing it.

        Args:
            key: Key of the record to retrieve

        Returns:
            The requested record, or None, if the record does not exist
        """

    @abstractmethod
    async def stream_record(self, key: str) -> AsyncContextManager[KeyValueStoreRecord[Response] | None]:
        """Retrieve the given record from the key-value store, as a stream.

        Args:
            key: Key of the record to retrieve

        Returns:
            The requested record as a context-managed streaming Response, or None, if the record does not exist
        """

    @abstractmethod
    async def set_record(self, key: str, value: Any, content_type: str | None = None) -> None:
        """Set a value to the given record in the key-value store.

        Args:
            key: The key of the record to save the value to
            value: The value to save into the record
            content_type: The content type of the saved value
        """

    @abstractmethod
    async def delete_record(self, key: str) -> None:
        """Delete the specified record from the key-value store.

        Args:
            key: The key of the record which to delete
        """



from __future__ import annotations

import csv
import io
import json
from typing import TYPE_CHECKING, AsyncIterator, Literal, TextIO, TypedDict, cast

from typing_extensions import NotRequired, Required, Unpack, override

from crawlee._utils.byte_size import ByteSize
from crawlee._utils.file import json_dumps
from crawlee.models import DatasetMetadata
from crawlee.storages.base_storage import BaseStorage
from crawlee.storages.key_value_store import KeyValueStore

if TYPE_CHECKING:
    from crawlee.base_storage_client import BaseStorageClient
    from crawlee.configuration import Configuration
    from crawlee.models import DatasetItemsListPage
    from crawlee.types import JSONSerializable


class GetDataKwargs(TypedDict):
    """Keyword arguments for dataset's `get_data` method.

    Args:
        offset: Skips the specified number of items at the start.
        limit: The maximum number of items to retrieve. Unlimited if None.
        clean: Returns only non-empty items and excludes hidden fields. Shortcut for skip_hidden and skip_empty.
        desc: Set True to sort results in descending order.
        fields: Fields to include in each item. Sorts fields as specified if provided.
        omit: Fields to exclude from each item.
        unwind: Unwinds items by a specified array field, turning each element into a separate item.
        skip_empty: Excludes empty items from the results if True.
        skip_hidden: Excludes fields starting with '#' if True.
        flatten: Fields to be flattened in returned items.
        view: Specifies the dataset view to be used.
    """

    offset: NotRequired[int]
    limit: NotRequired[int]
    clean: NotRequired[bool]
    desc: NotRequired[bool]
    fields: NotRequired[list[str]]
    omit: NotRequired[list[str]]
    unwind: NotRequired[str]
    skip_empty: NotRequired[bool]
    skip_hidden: NotRequired[bool]
    flatten: NotRequired[list[str]]
    view: NotRequired[str]


class PushDataKwargs(TypedDict):
    """Keyword arguments for dataset's `push_data` method."""


class ExportToKwargs(TypedDict):
    """Keyword arguments for dataset's `export_to` method.

    Args:
        key: The key under which to save the data.
        content_type: The format in which to export the data. Either 'json' or 'csv'.
        to_key_value_store_id: ID of the key-value store to save the exported file.
        to_key_value_store_name: Name of the key-value store to save the exported file.
    """

    key: Required[str]
    content_type: NotRequired[Literal['json', 'csv']]
    to_key_value_store_id: NotRequired[str]
    to_key_value_store_name: NotRequired[str]


class Dataset(BaseStorage):
    """Represents an append-only structured storage, ideal for tabular data akin to database tables.

    Represents a structured data store similar to a table, where each object (row) has consistent attributes (columns).
    Datasets operate on an append-only basis, allowing for the addition of new records without the modification or
    removal of existing ones. This class is typically used for storing crawling results.

    Data can be stored locally or in the cloud, with local storage paths formatted as:
    `{CRAWLEE_STORAGE_DIR}/datasets/{DATASET_ID}/{INDEX}.json`. Here, `{DATASET_ID}` is either "default" or
    a specific dataset ID, and `{INDEX}` represents the zero-based index of the item in the dataset.

    To open a dataset, use the `open` class method with an `id`, `name`, or `config`. If unspecified, the default
    dataset for the current crawler run is used. Opening a non-existent dataset by `id` raises an error, while
    by `name`, it is created.

    Usage:
        dataset = await Dataset.open(id='my_dataset_id')
    """

    _MAX_PAYLOAD_SIZE = ByteSize.from_mb(9)
    """Maximum size for a single payload."""

    _SAFETY_BUFFER_PERCENT = 0.01 / 100  # 0.01%
    """Percentage buffer to reduce payload limit slightly for safety."""

    _EFFECTIVE_LIMIT_SIZE = _MAX_PAYLOAD_SIZE - (_MAX_PAYLOAD_SIZE * _SAFETY_BUFFER_PERCENT)
    """Calculated payload limit considering safety buffer."""

    def __init__(
        self,
        id: str,
        name: str | None,
        configuration: Configuration,
        client: BaseStorageClient,
    ) -> None:
        self._id = id
        self._name = name
        self._configuration = configuration

        # Get resource clients from storage client
        self._resource_client = client.dataset(self._id)
        self._resource_collection_client = client.datasets()

    @override
    @property
    def id(self) -> str:
        return self._id

    @override
    @property
    def name(self) -> str | None:
        return self._name

    @override
    @classmethod
    async def open(
        cls,
        *,
        id: str | None = None,
        name: str | None = None,
        configuration: Configuration | None = None,
    ) -> Dataset:
        from crawlee.storages._creation_management import open_storage

        return await open_storage(
            storage_class=cls,
            id=id,
            name=name,
            configuration=configuration,
        )

    @override
    async def drop(self) -> None:
        from crawlee.storages._creation_management import remove_storage_from_cache

        await self._resource_client.delete()
        remove_storage_from_cache(storage_class=self.__class__, id=self._id, name=self._name)

    async def push_data(self, data: JSONSerializable, **kwargs: Unpack[PushDataKwargs]) -> None:
        """Store an object or an array of objects to the dataset.

        The size of the data is limited by the receiving API and therefore `push_data()` will only
        allow objects whose JSON representation is smaller than 9MB. When an array is passed,
        none of the included objects may be larger than 9MB, but the array itself may be of any size.

        Args:
            data: A JSON serializable data structure to be stored in the dataset. The JSON representation
                of each item must be smaller than 9MB.
            kwargs: Keyword arguments for the storage client method.
        """
        # Handle singular items
        if not isinstance(data, list):
            items = await self._check_and_serialize(data)
            return await self._resource_client.push_items(items, **kwargs)

        # Handle lists
        payloads_generator = (await self._check_and_serialize(item, index) for index, item in enumerate(data))

        # Invoke client in series to preserve the order of data
        async for items in self._chunk_by_size(payloads_generator):
            await self._resource_client.push_items(items, **kwargs)

        return None

    async def get_data(self, **kwargs: Unpack[GetDataKwargs]) -> DatasetItemsListPage:
        """Retrieves dataset items based on filtering, sorting, and pagination parameters.

        This method allows customization of the data retrieval process from a dataset, supporting operations such as
        field selection, ordering, and skipping specific records based on provided parameters.

        Args:
            kwargs: Keyword arguments for the storage client method.

        Returns:
            List page containing filtered and paginated dataset items.
        """
        # TODO: Improve error handling here
        # https://github.com/apify/apify-sdk-python/issues/140
        return await self._resource_client.list_items(**kwargs)

    async def write_to(self, content_type: Literal['json', 'csv'], destination: TextIO) -> None:
        """Exports the entire dataset into an arbitrary stream.

        Args:
            content_type: Specifies the output format
            destination: The stream into which the dataset contents should be written
        """
        items: list[dict] = []
        limit = 1000
        offset = 0

        while True:
            list_items = await self._resource_client.list_items(limit=limit, offset=offset)
            items.extend(list_items.items)
            if list_items.total <= offset + list_items.count:
                break
            offset += list_items.count

        if content_type == 'csv':
            writer = csv.writer(destination, quoting=csv.QUOTE_MINIMAL)
            writer.writerows([items[0].keys(), *[item.values() for item in items]])
        elif content_type == 'json':
            json.dump(items, destination)
        else:
            raise ValueError(f'Unsupported content type: {content_type}')

    async def export_to(self, **kwargs: Unpack[ExportToKwargs]) -> None:
        """Exports the entire dataset into a specified file stored under a key in a key-value store.

        This method consolidates all entries from a specified dataset into one file, which is then saved under a
        given key in a key-value store. The format of the exported file is determined by the `content_type` parameter.
        Either the dataset's ID or name should be specified, and similarly, either the target key-value store's ID or
        name should be used.

        Args:
            kwargs: Keyword arguments for the storage client method.
        """
        key = cast(str, kwargs.get('key'))
        content_type = kwargs.get('content_type', 'json')
        to_key_value_store_id = kwargs.get('to_key_value_store_id', None)
        to_key_value_store_name = kwargs.get('to_key_value_store_name', None)

        key_value_store = await KeyValueStore.open(id=to_key_value_store_id, name=to_key_value_store_name)

        output = io.StringIO()
        await self.write_to(content_type, output)

        if content_type == 'csv':
            await key_value_store.set_value(key, output.getvalue(), 'text/csv')

        if content_type == 'json':
            await key_value_store.set_value(key, output.getvalue(), 'application/json')

    async def get_info(self) -> DatasetMetadata | None:
        """Get an object containing general information about the dataset."""
        metadata = await self._resource_client.get()
        if isinstance(metadata, DatasetMetadata):
            return metadata
        return None

    async def iterate_items(
        self,
        *,
        offset: int = 0,
        limit: int | None = None,
        clean: bool = False,
        desc: bool = False,
        fields: list[str] | None = None,
        omit: list[str] | None = None,
        unwind: str | None = None,
        skip_empty: bool = False,
        skip_hidden: bool = False,
    ) -> AsyncIterator[dict]:
        """Iterates over dataset items, applying filtering, sorting, and pagination.

        Retrieves dataset items incrementally, allowing fine-grained control over the data fetched. The function
        supports various parameters to filter, sort, and limit the data returned, facilitating tailored dataset
        queries.

        Args:
            offset: Initial number of items to skip.
            limit: Max number of items to return. No limit if None.
            clean: Filters out empty items and hidden fields if True.
            desc: Returns items in reverse order if True.
            fields: Specific fields to include in each item.
            omit: Fields to omit from each item.
            unwind: Field name to unwind items by.
            skip_empty: Omits empty items if True.
            skip_hidden: Excludes fields starting with '#' if True.

        Yields:
            Each item from the dataset as a dictionary.
        """
        async for item in self._resource_client.iterate_items(  # type: ignore
            offset=offset,
            limit=limit,
            clean=clean,
            desc=desc,
            fields=fields,
            omit=omit,
            unwind=unwind,
            skip_empty=skip_empty,
            skip_hidden=skip_hidden,
        ):
            yield item

    async def _check_and_serialize(self, item: JSONSerializable, index: int | None = None) -> str:
        """Serializes a given item to JSON, checks its serializability and size against a limit.

        Args:
            item: The item to serialize.
            index: Index of the item, used for error context.

        Returns:
            Serialized JSON string.

        Raises:
            ValueError: If item is not JSON serializable or exceeds size limit.
        """
        s = ' ' if index is None else f' at index {index} '

        try:
            payload = await json_dumps(item)
        except Exception as exc:
            raise ValueError(f'Data item{s}is not serializable to JSON.') from exc

        payload_size = ByteSize(len(payload.encode('utf-8')))
        if payload_size > self._EFFECTIVE_LIMIT_SIZE:
            raise ValueError(f'Data item{s}is too large (size: {payload_size}, limit: {self._EFFECTIVE_LIMIT_SIZE})')

        return payload

    async def _chunk_by_size(self, items: AsyncIterator[str]) -> AsyncIterator[str]:
        """Yields chunks of JSON arrays composed of input strings, respecting a size limit.

        Groups an iterable of JSON string payloads into larger JSON arrays, ensuring the total size
        of each array does not exceed `EFFECTIVE_LIMIT_SIZE`. Each output is a JSON array string that
        contains as many payloads as possible without breaching the size threshold, maintaining the
        order of the original payloads. Assumes individual items are below the size limit.

        Args:
            items: Iterable of JSON string payloads.

        Yields:
            Strings representing JSON arrays of payloads, each staying within the size limit.
        """
        last_chunk_size = ByteSize(2)  # Add 2 bytes for [] wrapper.
        current_chunk = []

        async for payload in items:
            payload_size = ByteSize(len(payload.encode('utf-8')))

            if last_chunk_size + payload_size <= self._EFFECTIVE_LIMIT_SIZE:
                current_chunk.append(payload)
                last_chunk_size += payload_size + ByteSize(1)  # Add 1 byte for ',' separator.
            else:
                yield f'[{",".join(current_chunk)}]'
                current_chunk = [payload]
                last_chunk_size = payload_size + ByteSize(2)  # Add 2 bytes for [] wrapper.

        yield f'[{",".join(current_chunk)}]'



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from crawlee.basic_crawler.types import BasicCrawlingContext, EnqueueLinksFunction
from crawlee.http_crawler import HttpCrawlingResult

if TYPE_CHECKING:
    from bs4 import BeautifulSoup


@dataclass(frozen=True)
class BeautifulSoupCrawlingContext(HttpCrawlingResult, BasicCrawlingContext):
    """Crawling context used by BeautifulSoupCrawler."""

    soup: BeautifulSoup
    enqueue_links: EnqueueLinksFunction



from .models import FinalStatistics, StatisticsPersistedState, StatisticsState
from .statistics import Statistics

__all__ = ['FinalStatistics', 'Statistics', 'StatisticsPersistedState', 'StatisticsState']
